from __future__ import annotations

from abc import abstractmethod
from datetime import date, datetime, time
from decimal import Decimal
from typing import Any, Callable, List, Optional, Tuple, Union

from watchmen_auth import PrincipalService
from watchmen_data_kernel.common import ask_all_date_formats, ask_time_formats, DataKernelException
from watchmen_data_kernel.meta import TopicService
from watchmen_data_kernel.utils import MightAVariable, parse_function_in_variable, parse_variable
from watchmen_model.admin import Conditional, Factor, Topic
from watchmen_model.common import ComputedParameter, ConstantParameter, Parameter, ParameterComputeType, \
	ParameterCondition, ParameterExpression, ParameterExpressionOperator, ParameterJoint, ParameterJointType, \
	TopicFactorParameter, VariablePredefineFunctions
from watchmen_utilities import ArrayHelper, get_current_time_in_seconds, get_day_of_month, get_day_of_week, \
	get_half_year, get_month, get_quarter, get_week_of_month, get_week_of_year, get_year, greater_or_equals_date, \
	greater_or_equals_decimal, greater_or_equals_time, is_blank, is_date, is_date_or_time_instance, is_empty, \
	is_not_empty, is_numeric_instance, less_or_equals_date, less_or_equals_decimal, less_or_equals_time, \
	translate_date_format_to_memory, try_to_decimal, value_equals, value_not_equals
from .utils import always_none, compute_date_diff, create_from_previous_trigger_data, \
	create_get_from_variables_with_prefix, create_previous_trigger_data, create_snowflake_generator, \
	create_static_str, get_date_from_variables, get_value_from, test_date
from .variables import PipelineVariables


def get_topic_service(principal_service: PrincipalService) -> TopicService:
	return TopicService(principal_service)


# noinspection PyUnusedLocal
def always_true(variables: PipelineVariables, principal_service: PrincipalService) -> bool:
	return True


class ParsedMemoryCondition:
	def __init__(self, condition: ParameterCondition, principal_service: PrincipalService):
		self.condition = condition
		self.parse(condition, principal_service)

	@abstractmethod
	def parse(self, condition: ParameterCondition, principal_service: PrincipalService) -> None:
		pass

	@abstractmethod
	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> bool:
		pass


class ParsedMemoryParameter:
	def __init__(self, parameter: Optional[Parameter], principal_service: PrincipalService):
		self.parameter = parameter
		self.parse(parameter, principal_service)

	@abstractmethod
	def parse(self, parameter: Parameter, principal_service: PrincipalService) -> None:
		pass

	@abstractmethod
	def value(self, variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		pass


class NoopMemoryParameter(ParsedMemoryParameter):
	def parse(self, parameter: Parameter, principal_service: PrincipalService) -> None:
		"""
		do nothing
		"""
		pass

	def value(self, variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		"""
		always returns none
		"""
		return None


def parse_condition_in_memory(
		condition: Optional[ParameterCondition], principal_service: PrincipalService) -> ParsedMemoryCondition:
	if condition is None:
		raise DataKernelException('Condition cannot be none.')
	if isinstance(condition, ParameterJoint):
		return ParsedMemoryJoint(condition, principal_service)
	elif isinstance(condition, ParameterExpression):
		return ParsedMemoryExpression(condition, principal_service)
	else:
		raise DataKernelException(f'Condition[{condition.dict()}] is not supported.')


def parse_parameter_in_memory(
		parameter: Optional[Parameter], principal_service: PrincipalService) -> ParsedMemoryParameter:
	if parameter is None:
		return NoopMemoryParameter(None, principal_service)
	elif isinstance(parameter, TopicFactorParameter):
		return ParsedMemoryTopicFactorParameter(parameter, principal_service)
	elif isinstance(parameter, ConstantParameter):
		return ParsedMemoryConstantParameter(parameter, principal_service)
	elif isinstance(parameter, ComputedParameter):
		return ParsedMemoryComputedParameter(parameter, principal_service)
	else:
		raise DataKernelException(f'Parameter[{parameter.dict()}] is not supported.')


class ParsedMemoryJoint(ParsedMemoryCondition):
	jointType: ParameterJointType = ParameterJointType.AND
	filters: List[ParsedMemoryCondition] = []

	def parse(self, condition: ParameterJoint, principal_service: PrincipalService) -> None:
		self.jointType = ParameterJointType.OR \
			if condition.jointType == ParameterJointType.OR else ParameterJointType.AND
		self.filters = ArrayHelper(condition.filters) \
			.map(lambda x: parse_condition_in_memory(x, principal_service)).to_list()

	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> bool:
		if self.jointType == ParameterJointType.OR:
			return ArrayHelper(self.filters).some(lambda x: x.run(variables, principal_service))
		else:
			# and or not given
			return ArrayHelper(self.filters).every(lambda x: x.run(variables, principal_service))


class ParsedMemoryExpression(ParsedMemoryCondition):
	left: Optional[ParsedMemoryParameter] = None
	operator: Optional[ParameterExpressionOperator] = None
	right: Optional[ParsedMemoryParameter] = None

	def parse(self, condition: ParameterExpression, principal_service: PrincipalService) -> None:
		self.left = parse_parameter_in_memory(condition.left, principal_service)
		self.operator = condition.operator
		if self.operator == ParameterExpressionOperator.EMPTY or self.operator == ParameterExpressionOperator.NOT_EMPTY:
			# there is no need to parse right since it is unnecessary
			self.right = None
		else:
			self.right = parse_parameter_in_memory(condition.right, principal_service)

	def raise_cannot_compare(self, one: Any, another: Any) -> None:
		raise DataKernelException(
			f'Comparison of [none|int|float|decimal|date|time|datetime] are supported, '
			f'current are [one={one}, another={another}].')

	# noinspection PyMethodMayBeStatic
	def equals(self, one: Any, another: Any) -> bool:
		return value_equals(one, another, ask_time_formats(), ask_all_date_formats())

	# noinspection PyMethodMayBeStatic
	def not_equals(self, one: Any, another: Any) -> bool:
		return value_not_equals(one, another, ask_time_formats(), ask_all_date_formats())

	# noinspection PyMethodMayBeStatic
	def try_compare(self, func: Callable[[], Tuple[bool, bool]], or_raise: Callable[[], None]) -> bool:
		parsed, result = func()
		if parsed:
			return result
		else:
			or_raise()

	# noinspection PyMethodMayBeStatic
	def less_than(self, one: Any, another: Any) -> bool:
		if one is None:
			if another is None:
				return False
			elif is_numeric_instance(another) or is_date_or_time_instance(another):
				return True
		elif another is None:
			if is_numeric_instance(one) or is_date_or_time_instance(one):
				return False
		elif isinstance(one, int) or isinstance(one, float) or isinstance(one, Decimal):
			return self.try_compare(
				lambda: less_or_equals_decimal(one, another, False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, int) or isinstance(another, float) or isinstance(another, Decimal):
			return self.try_compare(
				lambda: greater_or_equals_decimal(one, another, True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(one, time):
			# compare time
			return self.try_compare(
				lambda: less_or_equals_time(one, another, ask_time_formats(), False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, time):
			# compare time
			return self.try_compare(
				lambda: greater_or_equals_time(another, one, ask_time_formats(), True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(one, datetime) or isinstance(one, date):
			# compare datetime or date
			return self.try_compare(
				lambda: less_or_equals_date(another, one, ask_all_date_formats(), False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, datetime) or isinstance(another, date):
			# compare datetime or date
			return self.try_compare(
				lambda: greater_or_equals_date(another, one, ask_all_date_formats(), True),
				lambda: self.raise_cannot_compare(one, another))

		self.raise_cannot_compare(one, another)

	# noinspection PyMethodMayBeStatic
	def less_than_or_equals(self, one: Any, another: Any) -> bool:
		if one is None:
			if another is None or is_numeric_instance(another) or is_date_or_time_instance(another):
				return True
		elif another is None:
			if is_numeric_instance(one) or is_date_or_time_instance(one):
				return False
		elif isinstance(one, int) or isinstance(one, float) or isinstance(one, Decimal):
			return self.try_compare(
				lambda: less_or_equals_decimal(one, another, True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, int) or isinstance(another, float) or isinstance(another, Decimal):
			return self.try_compare(
				lambda: greater_or_equals_decimal(one, another, False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(one, time):
			# compare time
			return self.try_compare(
				lambda: less_or_equals_time(one, another, ask_time_formats(), True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, time):
			# compare time
			return self.try_compare(
				lambda: greater_or_equals_time(another, one, ask_time_formats(), False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(one, datetime) or isinstance(one, date):
			# compare datetime or date
			return self.try_compare(
				lambda: less_or_equals_date(another, one, ask_all_date_formats(), True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, datetime) or isinstance(another, date):
			# compare datetime or date
			return self.try_compare(
				lambda: greater_or_equals_date(another, one, ask_all_date_formats(), False),
				lambda: self.raise_cannot_compare(one, another))

		self.raise_cannot_compare(one, another)

	# noinspection PyMethodMayBeStatic
	def greater_than(self, one: Any, another: Any) -> bool:
		if one is None:
			if another is None or is_numeric_instance(another) or is_date_or_time_instance(another):
				return False
		elif another is None:
			if is_numeric_instance(one) or is_date_or_time_instance(one):
				return True
		elif isinstance(one, int) or isinstance(one, float) or isinstance(one, Decimal):
			return self.try_compare(
				lambda: greater_or_equals_decimal(one, another, False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, int) or isinstance(another, float) or isinstance(another, Decimal):
			return self.try_compare(
				lambda: less_or_equals_decimal(one, another, True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(one, time):
			# compare time
			return self.try_compare(
				lambda: greater_or_equals_time(one, another, ask_time_formats(), False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, time):
			# compare time
			return self.try_compare(
				lambda: less_or_equals_time(another, one, ask_time_formats(), True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(one, datetime) or isinstance(one, date):
			# compare datetime or date
			return self.try_compare(
				lambda: greater_or_equals_date(another, one, ask_all_date_formats(), False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, datetime) or isinstance(another, date):
			# compare datetime or date
			return self.try_compare(
				lambda: less_or_equals_date(another, one, ask_all_date_formats(), True),
				lambda: self.raise_cannot_compare(one, another))

		self.raise_cannot_compare(one, another)

	# noinspection PyMethodMayBeStatic
	def greater_than_or_equals(self, one: Any, another: Any) -> bool:
		if one is None:
			if another is None:
				return True
			elif is_numeric_instance(another) or is_date_or_time_instance(another):
				return False
		elif another is None:
			if is_numeric_instance(one) or is_date_or_time_instance(one):
				return True
		elif isinstance(one, int) or isinstance(one, float) or isinstance(one, Decimal):
			return self.try_compare(
				lambda: greater_or_equals_decimal(one, another, True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, int) or isinstance(another, float) or isinstance(another, Decimal):
			return self.try_compare(
				lambda: less_or_equals_decimal(one, another, False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(one, time):
			# compare time
			return self.try_compare(
				lambda: greater_or_equals_time(one, another, ask_time_formats(), True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, time):
			# compare time
			return self.try_compare(
				lambda: less_or_equals_time(another, one, ask_time_formats(), False),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(one, datetime) or isinstance(one, date):
			# compare datetime or date
			return self.try_compare(
				lambda: greater_or_equals_date(another, one, ask_all_date_formats(), True),
				lambda: self.raise_cannot_compare(one, another))
		elif isinstance(another, datetime) or isinstance(another, date):
			# compare datetime or date
			return self.try_compare(
				lambda: less_or_equals_date(another, one, ask_all_date_formats(), False),
				lambda: self.raise_cannot_compare(one, another))

		self.raise_cannot_compare(one, another)

	# noinspection PyMethodMayBeStatic
	def exists(self, one: Any, another: Any) -> bool:
		if another is None:
			return False
		elif isinstance(another, list):
			return ArrayHelper(another).some(lambda x: self.equals(x, one))
		elif isinstance(another, str):
			if is_blank(another):
				return False
			return ArrayHelper(another.split(',')).some(lambda x: self.equals(x, one))
		else:
			raise DataKernelException(
				f'Comparison of [none|int|float|decimal|date|time|datetime] in [list|comma joined string] are supported, '
				f'current are [one={one}, another={another}].')

	def not_exists(self, one: Any, another: Any) -> bool:
		return not self.exists(one, another)

	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> bool:
		left_value = self.left.value(variables, principal_service)
		if self.operator == ParameterExpressionOperator.EMPTY:
			return is_empty(left_value)
		elif self.operator == ParameterExpressionOperator.NOT_EMPTY:
			return is_not_empty(left_value)

		right_value = self.right.value(variables, principal_service)
		if self.operator == ParameterExpressionOperator.EQUALS:
			return self.equals(left_value, right_value)
		elif self.operator == ParameterExpressionOperator.NOT_EQUALS:
			return self.not_equals(left_value, right_value)
		elif self.operator == ParameterExpressionOperator.LESS:
			return self.less_than(left_value, right_value)
		elif self.operator == ParameterExpressionOperator.LESS_EQUALS:
			return self.less_than_or_equals(left_value, right_value)
		elif self.operator == ParameterExpressionOperator.MORE:
			return self.greater_than(left_value, right_value)
		elif self.operator == ParameterExpressionOperator.MORE_EQUALS:
			return self.greater_than_or_equals(left_value, right_value)
		elif self.operator == ParameterExpressionOperator.IN:
			return self.exists(left_value, right_value)
		elif self.operator == ParameterExpressionOperator.NOT_IN:
			return self.not_exists(left_value, right_value)
		else:
			raise DataKernelException(
				f'Operator[{self.operator}] is not supported, found from expression[{self.condition.dict()}].')


def create_value_getter_from_current_data(name: str) -> Callable[[PipelineVariables, PrincipalService], Any]:
	return lambda variables, principal_service: variables.find_from_current_data(name)


def create_value_recursive_getter_from_current_data(
		name: str, names: List[str]
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	return lambda variables, principal_service: \
		get_value_from(name, names, lambda x: variables.find_from_current_data(x), variables.is_list_on_trigger)


def create_ask_factor_value(topic: Topic, factor: Factor) -> Callable[[PipelineVariables, PrincipalService], Any]:
	name = factor.name
	if is_blank(name):
		raise DataKernelException(f'Name of factor[id={factor.factorId}, topicId={topic.topicId}] not declared.')
	names = name.strip().split('.')

	# topic factor parameter always retrieve data from current trigger data
	if len(names) == 1:
		return create_value_getter_from_current_data(names[0])
	else:
		return create_value_recursive_getter_from_current_data(name, names)


class ParsedMemoryTopicFactorParameter(ParsedMemoryParameter):
	topic: Topic = None
	factor: Factor = None
	askValue: Callable[[PipelineVariables, PrincipalService], Any] = None

	def parse(self, parameter: TopicFactorParameter, principal_service: PrincipalService) -> None:
		if is_blank(parameter.topicId):
			raise DataKernelException(f'Topic not declared.')
		topic_service = get_topic_service(principal_service)
		topic: Optional[Topic] = topic_service.find_by_id(parameter.topicId)
		if topic is None:
			raise DataKernelException(f'Topic[id={parameter.topicId}] not found.')
		self.topic = topic

		if is_blank(parameter.factorId):
			raise DataKernelException(f'Factor not declared.')
		factor: Optional[Factor] = ArrayHelper(topic.factors).find(lambda x: x.factorId == parameter.factorId)
		if factor is None:
			raise DataKernelException(
				f'Factor[id={parameter.factorId}] in topic[id={topic.topicId}, name={topic.name}] not found.')
		self.factor = factor
		self.askValue = create_ask_factor_value(topic, factor)

	def value(self, variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		return self.askValue(variables, principal_service)


# noinspection DuplicatedCode
def create_date_diff(
		prefix: str, variable_name: str, function: VariablePredefineFunctions
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	parsed_params = parse_function_in_variable(variable_name, function.value, 2)
	end_variable_name = parsed_params[0]
	start_variable_name = parsed_params[1]
	end_parsed, end_date = test_date(end_variable_name)
	start_parsed, start_date = test_date(start_variable_name)
	if end_parsed and start_parsed:
		# noinspection PyUnusedLocal
		def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
			return compute_date_diff(function, end_date, start_date, variable_name)
	else:
		# noinspection DuplicatedCode
		def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
			if not end_parsed:
				e_parsed, e_value, e_date = get_date_from_variables(variables, principal_service, end_variable_name)
				if not e_parsed:
					raise DataKernelException(f'Value[{e_value}] cannot be parsed to date or datetime.')
			else:
				e_date = end_date
			if not start_parsed:
				s_parsed, s_value, s_date = get_date_from_variables(variables, principal_service, start_variable_name)
				if not s_parsed:
					raise DataKernelException(f'Value[{s_value}] cannot be parsed to date or datetime.')
			else:
				s_date = start_date
			return compute_date_diff(function, e_date, s_date, variable_name)

	def run(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		value = action(variables, principal_service)
		return value if is_blank(prefix) else f'{prefix}{value}'

	return run


def create_date_format(prefix: str, variable_name: str) -> Callable[[PipelineVariables, PrincipalService], Any]:
	parsed_params = parse_function_in_variable(variable_name, VariablePredefineFunctions.DATE_FORMAT.value, 2)
	variable_name = parsed_params[0]
	date_format = parsed_params[1]
	if is_blank(date_format):
		raise DataKernelException(f'Date format[{date_format}] cannot be recognized.')
	date_format = translate_date_format_to_memory(date_format)
	parsed, parsed_date = test_date(variable_name)
	if parsed:
		# noinspection PyUnusedLocal
		def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
			return parsed_date.strftime(date_format)
	else:
		def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
			date_parsed, value, a_date = get_date_from_variables(variables, principal_service, variable_name)
			if not date_parsed:
				raise DataKernelException(f'Value[{value}] cannot be parsed to date or datetime.')
			return a_date.strftime(date_format)

	def run(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		value = action(variables, principal_service)
		return value if is_blank(prefix) else f'{prefix}{value}'

	return run


# noinspection DuplicatedCode
def create_run_constant_segment(variable: MightAVariable) -> Callable[[PipelineVariables, PrincipalService], Any]:
	prefix = variable.text
	variable_name = variable.variable
	if variable_name == VariablePredefineFunctions.NEXT_SEQ.value:
		return create_snowflake_generator(prefix)
	elif variable_name == VariablePredefineFunctions.NOW.value:
		return lambda variables, principal_service: get_current_time_in_seconds()
	elif variable_name.startswith(VariablePredefineFunctions.YEAR_DIFF.value):
		return create_date_diff(prefix, variable_name, VariablePredefineFunctions.YEAR_DIFF)
	elif variable_name.startswith(VariablePredefineFunctions.MONTH_DIFF.value):
		return create_date_diff(prefix, variable_name, VariablePredefineFunctions.MONTH_DIFF)
	elif variable_name.startswith(VariablePredefineFunctions.DAY_DIFF.value):
		return create_date_diff(prefix, variable_name, VariablePredefineFunctions.DAY_DIFF)
	elif variable_name.startswith(VariablePredefineFunctions.DATE_FORMAT.value):
		return create_date_format(prefix, variable_name)
	elif variable_name.startswith(VariablePredefineFunctions.FROM_PREVIOUS_TRIGGER_DATA.value):
		if variable_name == VariablePredefineFunctions.FROM_PREVIOUS_TRIGGER_DATA.value:
			if is_blank(prefix):
				return create_previous_trigger_data()
			else:
				raise DataKernelException(f'Previous trigger data is a dict, cannot prefix by a string[{prefix}].')
		length = len(VariablePredefineFunctions.FROM_PREVIOUS_TRIGGER_DATA.value)
		if len(variable_name) < length + 2 or variable_name[length:length + 1] != '.':
			raise DataKernelException(f'Constant[{variable_name}] is not supported.')
		return create_from_previous_trigger_data(prefix, variable_name[length + 1:])
	else:
		return create_get_from_variables_with_prefix(prefix, variable_name)


def create_run_constant_segments(
		functions: List[Callable[[PipelineVariables, PrincipalService], Any]]
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		return ArrayHelper(functions).map(lambda x: x(variables, principal_service)).join('')

	return action


def create_ask_constant_value(variables: List[MightAVariable]) -> Callable[[PipelineVariables, PrincipalService], Any]:
	if len(variables) == 1:
		if variables[0].has_variable():
			return create_run_constant_segment(variables[0])
		else:
			return create_static_str(variables[0].text)
	else:
		return create_run_constant_segments(
			ArrayHelper(variables).map(lambda x: create_run_constant_segment(x)).to_list())


class ParsedMemoryConstantParameter(ParsedMemoryParameter):
	askValue: Callable[[PipelineVariables, PrincipalService], Any] = None

	# noinspection DuplicatedCode
	def parse(self, parameter: ConstantParameter, principal_service: PrincipalService) -> None:
		value = parameter.value
		if value is None:
			self.askValue = always_none
		elif len(value) == 0:
			self.askValue = always_none
		elif is_blank(value):
			self.askValue = create_static_str(value)
		elif '{' not in value or '}' not in value:
			self.askValue = create_static_str(value)
		else:
			_, variables = parse_variable(value)
			self.askValue = create_ask_constant_value(variables)

	def value(self, variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		return self.askValue(variables, principal_service)


def reducer_add(one: Decimal, another: Decimal) -> Decimal:
	return one + another


def reducer_subtract(one: Decimal, another: Decimal) -> Decimal:
	return one - another


def reducer_multiply(one: Decimal, another: Decimal) -> Decimal:
	return one * another


def reducer_divide(one: Decimal, another: Decimal) -> Decimal:
	return one / another


def reducer_modulus(one: Decimal, another: Decimal) -> Decimal:
	return one % another


def parse_to_decimal(value: Any, fallback_value: Callable[[], Optional[Decimal]] = lambda x: None) -> Optional[Decimal]:
	decimal_value = try_to_decimal(value)
	return fallback_value() if decimal_value is None else decimal_value


def create_numeric_reducer(
		parameters: List[ParsedMemoryParameter], reduce_func: Callable[[Decimal, Decimal], Decimal], numeric_name: str,
		fallback_first: Callable[[], Optional[Decimal]] = lambda x: None,
		fallback_rest: Callable[[], Optional[Decimal]] = lambda x: None
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	def reduce(variables: PipelineVariables, principal_service: PrincipalService) -> Decimal:
		first_value = parameters[0].value(variables, principal_service)
		first_decimal_value = parse_to_decimal(first_value, fallback_first)
		if first_decimal_value is None:
			raise DataKernelException(
				f'{numeric_name} [value={first_value}, type={type(first_value)}] is not supported.')

		result_decimal_value = first_decimal_value

		rest_parameters = parameters[1:]
		for rest_parameter in rest_parameters:
			rest_value = rest_parameter.value(variables, principal_service)
			rest_decimal_value = parse_to_decimal(rest_value, fallback_rest)
			if rest_decimal_value is None:
				raise DataKernelException(
					f'{numeric_name} [value={rest_value}, type={type(rest_value)}] is not supported.')
			result_decimal_value = reduce_func(result_decimal_value, rest_decimal_value)
		return result_decimal_value

	return reduce


def create_datetime_func(
		parameter: ParsedMemoryParameter, func: Callable[[date], int]
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	def get_part_of_datetime(variables: PipelineVariables, principal_service: PrincipalService) -> Optional[int]:
		value = parameter.value(variables, principal_service)
		if value is None:
			return None
		if isinstance(value, date):
			return func(value)
		parsed, dt_value = is_date(value, ask_all_date_formats())
		if not parsed:
			raise DataKernelException(f'Cannot parse value[{value}] to datetime.')
		if dt_value is None:
			return None
		return func(dt_value)

	return get_part_of_datetime


def create_case_then(
		cases: List[Tuple[PrerequisiteTest, ParsedMemoryParameter]], anyway: Optional[ParsedMemoryParameter]
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	def run_case_then(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		found: Optional[Tuple[PrerequisiteTest, ParsedMemoryParameter]] = \
			ArrayHelper(cases).find(lambda x: x[0](variables, principal_service))
		if found is not None:
			# find a route
			return found[1].value(variables, principal_service)
		elif anyway is not None:
			# return anyway value when no route found
			return anyway.value(variables, principal_service)
		else:
			# return none when no route found and no anyway route
			return None

	return run_case_then


def parse_conditional_parameter_in_memory(
		parameter: Parameter, principal_service: PrincipalService
) -> Tuple[PrerequisiteTest, ParsedMemoryParameter]:
	return \
		parse_prerequisite_in_memory(parameter, principal_service), \
		parse_parameter_in_memory(parameter, principal_service)


def assert_parameter_count(
		func_name: str, parameters: Optional[List[Parameter]],
		min_count: int = 1, max_count: int = 9999, allow_undefined: bool = False
) -> None:
	if parameters is None:
		raise DataKernelException(f'Parameter not found on computation[{func_name}].')
	count = len(parameters)
	if count < min_count:
		raise DataKernelException(
			f'At least {min_count} parameter(s) on computation[{func_name}], current is [{parameters}].')
	if count > max_count:
		raise DataKernelException(
			f'At most {max_count} parameter(s) on computation[{func_name}], current is [{parameters}].')
	if not allow_undefined:
		found = ArrayHelper(parameters).some(lambda x: x is None)
		if found:
			raise DataKernelException(
				f'None parameter is not allowed on computation[{func_name}], current is [{parameters}].')


class ParsedMemoryComputedParameter(ParsedMemoryParameter):
	askValue: Callable[[PipelineVariables, PrincipalService], Any] = None

	def parse(self, parameter: ComputedParameter, principal_service: PrincipalService) -> None:
		compute_type = parameter.type
		if is_blank(compute_type) or compute_type == ParameterComputeType.NONE:
			raise DataKernelException(f'Compute type not declared.')

		def parse_parameter(param: Parameter) -> ParsedMemoryParameter:
			return parse_parameter_in_memory(param, principal_service)

		def parse_sub_parameters(param: ComputedParameter) -> List[ParsedMemoryParameter]:
			return ArrayHelper(param.parameters).map(parse_parameter).to_list()

		if compute_type == ParameterComputeType.ADD:
			assert_parameter_count('add', parameter.parameters, 2)
			# treat none value as 0
			self.askValue = create_numeric_reducer(
				parse_sub_parameters(parameter), reducer_add, 'Add', lambda: Decimal(0), lambda: Decimal(0))
		elif compute_type == ParameterComputeType.SUBTRACT:
			assert_parameter_count('subtract', parameter.parameters, 2)
			# treat none value as 0
			self.askValue = create_numeric_reducer(
				parse_sub_parameters(parameter), reducer_subtract, 'Subtract', lambda: Decimal(0), lambda: Decimal(0))
		elif compute_type == ParameterComputeType.MULTIPLY:
			assert_parameter_count('multiply', parameter.parameters, 2)
			self.askValue = create_numeric_reducer(parse_sub_parameters(parameter), reducer_multiply, 'Multiply')
		elif compute_type == ParameterComputeType.DIVIDE:
			assert_parameter_count('divide', parameter.parameters, 2)
			self.askValue = create_numeric_reducer(parse_sub_parameters(parameter), reducer_divide, 'Divide')
		elif compute_type == ParameterComputeType.MODULUS:
			assert_parameter_count('modulus', parameter.parameters, 2)
			self.askValue = create_numeric_reducer(parse_sub_parameters(parameter), reducer_modulus, 'Modulus')
		elif compute_type == ParameterComputeType.YEAR_OF:
			assert_parameter_count('year-of', parameter.parameters, 1, 1)
			self.askValue = create_datetime_func(parse_parameter(parameter.parameters[0]), get_year)
		elif compute_type == ParameterComputeType.HALF_YEAR_OF:
			assert_parameter_count('half-year-of', parameter.parameters, 1, 1)
			self.askValue = create_datetime_func(parse_parameter(parameter.parameters[0]), get_half_year)
		elif compute_type == ParameterComputeType.QUARTER_OF:
			assert_parameter_count('quarter-of', parameter.parameters, 1, 1)
			self.askValue = create_datetime_func(parse_parameter(parameter.parameters[0]), get_quarter)
		elif compute_type == ParameterComputeType.MONTH_OF:
			assert_parameter_count('month-of', parameter.parameters, 1, 1)
			self.askValue = create_datetime_func(parse_parameter(parameter.parameters[0]), get_month)
		elif compute_type == ParameterComputeType.WEEK_OF_YEAR:
			assert_parameter_count('week-of-year', parameter.parameters, 1, 1)
			self.askValue = create_datetime_func(parse_parameter(parameter.parameters[0]), get_week_of_year)
		elif compute_type == ParameterComputeType.WEEK_OF_MONTH:
			assert_parameter_count('week-of-month', parameter.parameters, 1, 1)
			self.askValue = create_datetime_func(parse_parameter(parameter.parameters[0]), get_week_of_month)
		elif compute_type == ParameterComputeType.DAY_OF_MONTH:
			assert_parameter_count('day-of-month', parameter.parameters, 1, 1)
			self.askValue = create_datetime_func(parse_parameter(parameter.parameters[0]), get_day_of_month)
		elif compute_type == ParameterComputeType.DAY_OF_WEEK:
			assert_parameter_count('day-of-week', parameter.parameters, 1, 1)
			self.askValue = create_datetime_func(parse_parameter(parameter.parameters[0]), get_day_of_week)
		elif compute_type == ParameterComputeType.CASE_THEN:
			# noinspection DuplicatedCode
			assert_parameter_count('case-then', parameter.parameters, 1)
			cases = parameter.parameters
			if cases is None or len(cases) == 0:
				raise DataKernelException(f'Case not declared in case then computation.')
			anyways = ArrayHelper(cases).filter(lambda x: not x.conditional).to_list()
			if len(anyways) > 1:
				raise DataKernelException(
					f'Multiple anyway routes declared in case then computation[{parameter.dict()}].')

			def parse_route(param: Parameter) -> Tuple[PrerequisiteTest, ParsedMemoryParameter]:
				return parse_conditional_parameter_in_memory(param, principal_service)

			# noinspection DuplicatedCode
			routes = ArrayHelper(cases).filter(lambda x: x.conditional).map(parse_route).to_list()
			anyway = anyways[0] if len(anyways) == 1 else None
			if anyway is not None:
				anyway_route = parse_parameter(anyway)
			else:
				anyway_route = None
			self.askValue = create_case_then(routes, anyway_route)
		else:
			raise DataKernelException(f'Compute type[{compute_type}] is not supported.')

	def value(self, variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		return self.askValue(variables, principal_service)


def ask_condition(
		condition: ParsedMemoryCondition, variables: PipelineVariables,
		principal_service: PrincipalService) -> bool:
	return condition.run(variables, principal_service)


PrerequisiteTest = Callable[[PipelineVariables, PrincipalService], bool]


def create_ask_prerequisite(condition: ParsedMemoryCondition) -> PrerequisiteTest:
	def ask(variables: PipelineVariables, principal_service: PrincipalService) -> bool:
		return ask_condition(condition, variables, principal_service)

	return ask


def parse_prerequisite_in_memory(
		conditional: Union[Conditional, Parameter], principal_service: PrincipalService) -> PrerequisiteTest:
	if conditional.conditional is None or not conditional.conditional:
		# no condition is needed
		return always_true

	joint = conditional.on
	if joint is None:
		# no condition defined
		return always_true

	filters = joint.filters
	if filters is None or len(filters) == 0:
		# no filters defined
		return always_true
	condition = ParsedMemoryJoint(joint, principal_service)

	return create_ask_prerequisite(condition)


PrerequisiteDefinedAs = Callable[[], Any]


# noinspection PyUnusedLocal
def parse_prerequisite_defined_as(
		conditional: Conditional, principal_service: PrincipalService
) -> PrerequisiteDefinedAs:
	defined_as = {
		'conditional': False if conditional.conditional is None else conditional.conditional,
		'on': None if conditional.on is None else conditional.on.dict()
	}

	return lambda: defined_as



from base64 import b64encode
from hashlib import md5
from typing import Any, Union

from watchmen_model.admin import FactorEncryptMethod
from .encryptor import Encryptor


class MD5Encryptor(Encryptor):
	def accept(self, method: Union[FactorEncryptMethod, str]) -> bool:
		return method == FactorEncryptMethod.MD5

	def is_encrypted(self, value: Any) -> bool:
		if value is None:
			return False
		elif isinstance(value, str) and value.startswith('{MD5}'):
			return True
		else:
			return False

	def do_encrypt(self, value: Any) -> Any:
		value = str(value)
		return b64encode(md5(value.encode('utf-8')).digest()).decode()

	def do_decrypt(self, value: str) -> str:
		# md5 cannot be decrypted, remove prefix only
		return value[5:]

	def get_key_type(self) -> str:
		return FactorEncryptMethod.MD5.value



from typing import Any, Callable, Hashable, Optional, Union, ValuesView

from cacheout import Cache

from watchmen_data_kernel.common import ask_cache_enabled


class InternalCache:
	def __init__(self, cache: Callable[[], Cache]):
		if ask_cache_enabled():
			self.cache = cache()

	def put(self, key: Hashable, value: Any, ttl: Optional[Union[int, float]] = None) -> Optional[Any]:
		if not ask_cache_enabled():
			return None

		existing: Optional[Any] = self.cache.get(key)
		self.cache.set(key, value, ttl)
		return existing

	def get(self, key: Hashable, default_value: Optional[Any] = None) -> Optional[Any]:
		if not ask_cache_enabled():
			return None

		existing: Optional[Any] = self.cache.get(key)
		return existing if existing is not None else default_value

	def remove(self, key: Hashable) -> Optional[Any]:
		if not ask_cache_enabled():
			return None

		existing: Optional[Any] = self.cache.get(key)
		self.cache.delete(key)
		return existing

	def values(self) -> ValuesView:
		return self.cache.values()

	def clear(self) -> None:
		if ask_cache_enabled():
			self.cache.clear()



from typing import Optional

from watchmen_auth import PrincipalService
from watchmen_data_kernel.cache import CacheService
from watchmen_data_kernel.common import DataKernelException
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_meta.system import DataSourceService as DataSourceStorageService
from watchmen_model.common import DataSourceId
from watchmen_model.system import DataSource


class DataSourceService:
	def __init__(self, principal_service: PrincipalService):
		self.principalService = principal_service

	def find_by_id(self, data_source_id: DataSourceId) -> Optional[DataSource]:
		data_source = CacheService.data_source().get(data_source_id)
		if data_source is not None:
			if data_source.tenantId != self.principalService.get_tenant_id():
				raise DataKernelException(
					f'Data source[id={data_source_id}] not belongs to '
					f'current tenant[id={self.principalService.get_tenant_id()}].')
			return data_source

		storage_service = DataSourceStorageService(
			ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			data_source: DataSource = storage_service.find_by_id(data_source_id)
			if data_source is None:
				return None

			CacheService.data_source().put(data_source)
			return data_source
		finally:
			storage_service.close_transaction()



from typing import List, Optional

from watchmen_auth import PrincipalService
from watchmen_data_kernel.cache import CacheService
from watchmen_data_kernel.common import DataKernelException
from watchmen_data_kernel.topic_schema import TopicSchema
from watchmen_meta.admin import TopicService as TopicStorageService
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_model.admin import Topic, TopicKind
from watchmen_model.common import TenantId, TopicId
from watchmen_utilities import ArrayHelper


class TopicService:
	def __init__(self, principal_service: PrincipalService):
		self.principalService = principal_service

	def find_by_id(self, topic_id: TopicId) -> Optional[Topic]:
		topic = CacheService.topic().get(topic_id)
		if topic is not None:
			if topic.tenantId != self.principalService.get_tenant_id():
				raise DataKernelException(
					f'Topic[id={topic_id}] not belongs to current tenant[id={self.principalService.get_tenant_id()}].')
			return topic

		storage_service = TopicStorageService(ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			topic: Topic = storage_service.find_by_id(topic_id)
			if topic is None:
				return None

			CacheService.topic().put(topic)
			return topic
		finally:
			storage_service.close_transaction()

	def find_schema_by_id(self, topic_id: TopicId, tenant_id: TenantId) -> Optional[TopicSchema]:
		if not self.principalService.is_super_admin():
			if self.principalService.get_tenant_id() != tenant_id:
				raise Exception('Forbidden')

		schema = CacheService.topic().get_schema(topic_id)
		if schema is not None:
			if schema.get_topic().tenantId != tenant_id:
				return None
			return schema

		storage_service = TopicStorageService(ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			topic: Topic = storage_service.find_by_id(topic_id)
			if topic is None:
				return None

			CacheService.topic().put(topic)
			schema = CacheService.topic().get_schema(topic.topicId)
			if schema is not None:
				if schema.get_topic().tenantId != tenant_id:
					return None
			return schema
		finally:
			storage_service.close_transaction()

	def find_schema_by_name(self, name: str, tenant_id: TenantId) -> Optional[TopicSchema]:
		if not self.principalService.is_super_admin():
			if self.principalService.get_tenant_id() != tenant_id:
				raise Exception('Forbidden')

		schema = None
		topic = CacheService.topic().get_by_name(name, tenant_id)
		if topic is not None:
			schema = CacheService.topic().get_schema(topic.topicId)
		if schema is not None:
			return schema

		storage_service = TopicStorageService(ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			topic: Topic = storage_service.find_by_name_and_tenant(name, tenant_id)
			if topic is None:
				return None

			CacheService.topic().put(topic)
			return CacheService.topic().get_schema(topic.topicId)
		finally:
			storage_service.close_transaction()

	def find_should_monitored(self, tenant_id: TenantId) -> List[Topic]:
		storage_service = TopicStorageService(ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			topics = storage_service.find_all(tenant_id)
			# only business topics need to be monitored
			return ArrayHelper(topics).filter(lambda x: x.kind == TopicKind.BUSINESS).to_list()
		finally:
			storage_service.close_transaction()



from typing import Union

from watchmen_auth import PrincipalService
from watchmen_data_kernel.cache import CacheService
from watchmen_data_kernel.common import DataKernelException
from watchmen_data_kernel.meta import DataSourceService
from watchmen_data_kernel.storage.topic_storage import build_topic_data_storage
from watchmen_data_kernel.topic_schema import TopicSchema
from watchmen_model.admin import Topic
from watchmen_storage import TopicDataStorageSPI
from watchmen_utilities import is_blank


def get_data_source_service(principal_service: PrincipalService) -> DataSourceService:
	return DataSourceService(principal_service)


def ask_topic_storage(
		topic_or_schema: Union[Topic, TopicSchema], principal_service: PrincipalService) -> TopicDataStorageSPI:
	topic = topic_or_schema if isinstance(topic_or_schema, Topic) else topic_or_schema.get_topic()
	data_source_id = topic.dataSourceId
	if is_blank(data_source_id):
		raise DataKernelException(
			f'Data source is not defined for topic[id={topic.topicId}, name={topic.name}]')

	build = CacheService.data_source().get_builder(data_source_id)
	if build is not None:
		return build()

	data_source = get_data_source_service(principal_service).find_by_id(data_source_id)
	if data_source is None:
		raise DataKernelException(
			f'Data source not declared for topic'
			f'[id={topic.topicId}, name={topic.name}, dataSourceId={data_source_id}]')

	build = build_topic_data_storage(data_source)
	CacheService.data_source().put_builder(data_source_id, build)
	return build()



from logging import getLogger
from typing import Dict, Optional, Union

from watchmen_data_kernel.common import ask_all_date_formats, ask_encrypt_aes_params
from watchmen_model.admin import FactorEncryptMethod
from .aes_encryptor import AESEncryptor
from .center_masker import CenterMasker
from .date_masker import DateMasker
from .encryptor import Encryptor
from .last_masker import LastMasker
from .mail_masker import MailMasker
from .md5_encryptor import MD5Encryptor
from .sha256_encryptor import SHA256Encryptor

logger = getLogger(__name__)


class EncryptorRegistry:
	encryptors: Dict[str, Encryptor] = {}

	# noinspection PyMethodMayBeStatic
	def to_key(self, method: Union[FactorEncryptMethod, str]) -> str:
		return method.value if isinstance(method, FactorEncryptMethod) else method

	def register(self, method: Union[FactorEncryptMethod, str], encryptor: Encryptor) -> Optional[Encryptor]:
		key = self.to_key(method)
		original = self.encryptors.get(key)
		self.encryptors[key] = encryptor
		if original is not None:
			logger.warning(f'Encryptor[{method}] is replaced.')
		else:
			logger.info(f'Encryptor[{method}] is registered.')
		return original

	def is_registered(self, method: Union[FactorEncryptMethod, str]) -> bool:
		key = self.to_key(method)
		return key in self.encryptors

	def ask_encryptor(self, method: Union[FactorEncryptMethod, str]) -> Optional[Encryptor]:
		key = self.to_key(method)
		return self.encryptors.get(key)


encryptor_registry = EncryptorRegistry()


def register_encryptor(method: Union[FactorEncryptMethod, str], encryptor: Encryptor) -> None:
	"""
	register writer on startup
	"""
	encryptor_registry.register(method, encryptor)


def is_encryptor_registered(method: Union[FactorEncryptMethod, str]) -> bool:
	return encryptor_registry.is_registered(method)


def find_encryptor(method: Union[FactorEncryptMethod, str]) -> Optional[Encryptor]:
	return encryptor_registry.ask_encryptor(method)


# register default
aes_key, aes_iv = ask_encrypt_aes_params()
register_encryptor(FactorEncryptMethod.AES256_PKCS5_PADDING, AESEncryptor(key=aes_key, iv=aes_iv))
register_encryptor(FactorEncryptMethod.MD5, MD5Encryptor())
register_encryptor(FactorEncryptMethod.SHA256, SHA256Encryptor())
register_encryptor(FactorEncryptMethod.MASK_MAIL, MailMasker())
register_encryptor(FactorEncryptMethod.MASK_CENTER_3, CenterMasker(3))
register_encryptor(FactorEncryptMethod.MASK_CENTER_5, CenterMasker(5))
register_encryptor(FactorEncryptMethod.MASK_LAST_3, LastMasker(3))
register_encryptor(FactorEncryptMethod.MASK_LAST_6, LastMasker(6))
date_formats = ask_all_date_formats()
register_encryptor(FactorEncryptMethod.MASK_MONTH_DAY, DateMasker(month=True, day=True, formats=date_formats))
register_encryptor(FactorEncryptMethod.MASK_MONTH, DateMasker(month=True, day=False, formats=date_formats))
register_encryptor(FactorEncryptMethod.MASK_DAY, DateMasker(month=False, day=True, formats=date_formats))



from __future__ import annotations

from typing import Any, Dict, List, Optional

from watchmen_data_kernel.common import ask_ignore_default_on_raw
from watchmen_model.admin import Factor, Topic
from watchmen_utilities import ArrayHelper, is_blank, is_not_blank
from .utils import cast_value_for_factor

"""
design of default value initializer. same with encryption, see that for more. 
"""


class DefaultValueFactor:
	def __init__(self, factor: Factor):
		self.factor = factor
		self.factorName = '' if is_blank(factor.name) else factor.name.strip()
		self.names = self.factorName.split('.')
		self.defaultValue = cast_value_for_factor(factor.defaultValue, factor)

	def pop_first_name(self):
		self.names = self.names[1:]

	def get_default_value(self):
		return self.defaultValue


class DefaultValueFactorGroup:
	factors: Optional[List[DefaultValueFactor]] = None
	groups: Optional[List[DefaultValueFactorGroup]] = None

	def __init__(self, name: str, factors: List[DefaultValueFactor]):
		self.name = name
		# in reality, zero or one factor.
		# if there is one, name is same as group's, and will not contain group anymore
		self.factors = ArrayHelper(factors).filter(lambda x: len(x.names) == 1).to_list()
		groups = ArrayHelper(factors).filter(lambda x: len(x.names) > 1) \
			.each(lambda x: x.pop_first_name()) \
			.group_by(lambda x: x.names[0])
		self.groups = ArrayHelper(list(groups.items())) \
			.map(lambda x: DefaultValueFactorGroup(name=x[0], factors=x[1])).to_list()

	def set_default_value(self, data: Dict[str, Any]) -> None:
		value = data.get(self.name)
		if isinstance(value, dict):
			ArrayHelper(self.groups).each(lambda x: x.set_default_value(value))
		elif isinstance(value, list):
			def each(item):
				ArrayHelper(self.groups).each(lambda x: x.set_default_value(item))

			ArrayHelper(value).each(lambda x: each(x))
		elif value is None and len(self.factors) == 1:
			data[self.name] = self.factors[0].get_default_value()


def parse_default_value_factors(topic: Topic) -> List[DefaultValueFactorGroup]:
	if ask_ignore_default_on_raw():
		return []

	groups = ArrayHelper(topic.factors) \
		.filter(lambda x: x.defaultValue is not None) \
		.map(lambda x: DefaultValueFactor(x)) \
		.filter(lambda x: is_not_blank(x.factorName)) \
		.group_by(lambda x: x.names[0])

	return ArrayHelper(list(groups.items())) \
		.map(lambda x: DefaultValueFactorGroup(name=x[0], factors=x[1])).to_list()



from logging import getLogger
from threading import Thread
from typing import Optional

from time import sleep

from watchmen_data_kernel.common import ask_cache_heart_beat_enabled, ask_cache_heart_beat_interval
from watchmen_meta.admin import PipelineService, TopicService
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator, ask_super_admin
from watchmen_meta.system import DataSourceService, ExternalWriterService, TenantService
from watchmen_model.admin import Pipeline, Topic
from watchmen_model.system import DataSource, ExternalWriter, Tenant
from .data_source_cache import data_source_cache, DataSourceCache
from .external_writer_cache import external_writer_cache, ExternalWriterCache
from .key_store_cache import key_store_cache, KeyStoreCache
from .pipeline_by_topic_cache import pipeline_by_topic_cache, PipelineByTopicCache
from .pipeline_cache import pipeline_cache, PipelineCache
from .tenant_cache import tenant_cache, TenantCache
from .topic_cache import topic_cache, TopicCache


class CacheService:
	@staticmethod
	def pipeline() -> PipelineCache:
		return pipeline_cache

	@staticmethod
	def topic() -> TopicCache:
		return topic_cache

	@staticmethod
	def pipelines_by_topic() -> PipelineByTopicCache:
		return pipeline_by_topic_cache

	@staticmethod
	def data_source() -> DataSourceCache:
		return data_source_cache

	@staticmethod
	def external_writer() -> ExternalWriterCache:
		return external_writer_cache

	@staticmethod
	def tenant() -> TenantCache:
		return tenant_cache

	@staticmethod
	def key_store() -> KeyStoreCache:
		return key_store_cache

	@staticmethod
	def clear_all() -> None:
		CacheService.pipeline().clear()
		CacheService.topic().clear()
		CacheService.pipelines_by_topic().clear()
		CacheService.data_source().clear()
		CacheService.external_writer().clear()
		CacheService.tenant().clear()
		CacheService.key_store().clear()


def heart_beat_on_pipelines() -> None:
	pipelines = CacheService.pipeline().all()
	pipeline_service = PipelineService(ask_meta_storage(), ask_snowflake_generator(), ask_super_admin())
	pipeline_service.begin_transaction()
	try:
		for pipeline in pipelines:
			loaded: Optional[Pipeline] = pipeline_service.find_by_id(pipeline.pipelineId)
			if loaded is None:
				CacheService.pipeline().remove(pipeline.pipelineId)
			elif loaded.lastModifiedAt > pipeline.lastModifiedAt or loaded.version > pipeline.version:
				CacheService.pipeline().put(loaded)
	finally:
		pipeline_service.close_transaction()


def heart_beat_on_topics() -> None:
	topics = CacheService.topic().all()
	topic_service = TopicService(ask_meta_storage(), ask_snowflake_generator(), ask_super_admin())
	topic_service.begin_transaction()
	try:
		for topic in topics:
			loaded: Optional[Topic] = topic_service.find_by_id(topic.topicId)
			if loaded is None:
				CacheService.topic().remove(topic.topicId)
			elif loaded.lastModifiedAt > topic.lastModifiedAt or loaded.version > topic.version:
				CacheService.topic().put(loaded)
	finally:
		topic_service.close_transaction()


def heart_beat_on_data_sources() -> None:
	data_sources = CacheService.data_source().all()
	data_source_service = DataSourceService(ask_meta_storage(), ask_snowflake_generator(), ask_super_admin())
	data_source_service.begin_transaction()
	try:
		for data_source in data_sources:
			loaded: Optional[DataSource] = data_source_service.find_by_id(data_source.dataSourceId)
			if loaded is None:
				CacheService.data_source().remove(data_source.dataSourceId)
			elif loaded.lastModifiedAt > data_source.lastModifiedAt or loaded.version > data_source.version:
				CacheService.data_source().put(loaded)
	finally:
		data_source_service.close_transaction()


def heart_beat_on_external_writers() -> None:
	external_writers = CacheService.external_writer().all()
	external_writer_service = ExternalWriterService(ask_meta_storage(), ask_snowflake_generator(), ask_super_admin())
	external_writer_service.begin_transaction()
	try:
		for external_writer in external_writers:
			loaded: Optional[ExternalWriter] = external_writer_service.find_by_id(external_writer.writerId)
			if loaded is None:
				CacheService.external_writer().remove(external_writer.writerId)
			elif loaded.lastModifiedAt > external_writer.lastModifiedAt or loaded.version > external_writer.version:
				CacheService.external_writer().put(loaded)
	finally:
		external_writer_service.close_transaction()


def heart_beat_on_tenants() -> None:
	tenants = CacheService.tenant().all()
	tenant_service = TenantService(ask_meta_storage(), ask_snowflake_generator(), ask_super_admin())
	tenant_service.begin_transaction()
	try:
		for tenant in tenants:
			loaded: Optional[Tenant] = tenant_service.find_by_id(tenant.tenantId)
			if loaded is None:
				CacheService.tenant().remove(tenant.tenantId)
			elif loaded.lastModifiedAt > tenant.lastModifiedAt or loaded.version > tenant.version:
				CacheService.tenant().put(loaded)
	finally:
		tenant_service.close_transaction()


# cache refresher, with heart beat
def cache_heart_beat():
	logger = getLogger(__name__)
	logger.info('Cache heart beat started.')
	interval = ask_cache_heart_beat_interval()
	try:
		while True:
			sleep(interval)
			# cache heart beat
			heart_beat_on_tenants()
			heart_beat_on_data_sources()
			heart_beat_on_external_writers()
			heart_beat_on_pipelines()
			heart_beat_on_topics()
	except Exception as e:
		logger.error(e, exc_info=True, stack_info=True)
	finally:
		logger.warning('Cache heart beat stopped.')
		# try to restart
		logger.info('Try to restart cache heart beat.')
		cache_heart_beat()


if ask_cache_heart_beat_enabled():
	Thread(target=cache_heart_beat, args=(), daemon=True).start()



from .external_writer_builder_registry import CreateExternalWriterBuilder, find_external_writer_create, \
	register_external_writer_create
from .external_writer_registry import ask_external_writer_creator, BuildExternalWriter, ExternalWriter, \
	ExternalWriterParams, is_external_writer_creator_registered, register_external_writer_creator



from typing import Any, Union

from watchmen_model.admin import FactorEncryptMethod
from .encryptor import Encryptor


class SHA256Encryptor(Encryptor):
	def accept(self, method: Union[FactorEncryptMethod, str]) -> bool:
		return method == FactorEncryptMethod.SHA256

	def is_encrypted(self, value: Any) -> bool:
		if value is None:
			return False
		elif isinstance(value, str) and value.startswith('{SHA256}'):
			return True
		else:
			return False

	# noinspection PyMethodMayBeStatic
	def ask_sha256(self):
		# lazy load
		# noinspection PyPackageRequirements
		from Crypto.Hash import SHA256
		return SHA256.new()

	def do_encrypt(self, value: Any) -> Any:
		value = str(value)
		sha256 = self.ask_sha256()
		sha256.update(value.encode('utf-8'))
		return sha256.hexdigest()

	def do_decrypt(self, value: str) -> str:
		# sha256 cannot be decrypted, remove prefix only
		return value[8:]

	def get_key_type(self) -> str:
		return FactorEncryptMethod.SHA256.value



from __future__ import annotations

from abc import abstractmethod
from datetime import date
from decimal import Decimal
from enum import Enum
from inspect import isfunction
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from watchmen_auth import PrincipalService
from watchmen_data_kernel.common import ask_all_date_formats, ask_time_formats, DataKernelException
from watchmen_data_kernel.meta import TopicService
from watchmen_data_kernel.topic_schema import cast_value_for_factor, TopicSchema
from watchmen_data_kernel.utils import MightAVariable, parse_function_in_variable, parse_variable
from watchmen_model.admin import AccumulateMode, AggregateArithmetic, Factor, FactorType, FromTopic, InsertRowAction, \
	is_raw_topic, MappingFactor, Topic, TopicKind, ToTopic
from watchmen_model.admin.pipeline_action_write import InsertOrMergeRowAction, WriteTopicAction
from watchmen_model.common import ComputedParameter, ConstantParameter, FactorId, Parameter, ParameterComputeType, \
	ParameterCondition, ParameterExpression, ParameterExpressionOperator, ParameterJoint, ParameterJointType, \
	ParameterKind, TopicFactorParameter, TopicId, VariablePredefineFunctions
from watchmen_model.pipeline_kernel import TopicDataColumnNames
from watchmen_storage import ColumnNameLiteral, ComputedLiteral, ComputedLiteralOperator, EntityCriteriaExpression, \
	EntityCriteriaJoint, EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityCriteriaStatement, Literal
from watchmen_utilities import ArrayHelper, get_current_time_in_seconds, is_blank, is_date, is_decimal, is_time, \
	translate_date_format_to_memory
from .ask_from_memory import assert_parameter_count, create_ask_factor_value, parse_parameter_in_memory
from .topic_utils import ask_topic_data_entity_helper
from .utils import always_none, compute_date_diff, create_from_previous_trigger_data, \
	create_get_from_variables_with_prefix, create_snowflake_generator, create_static_str, get_date_from_variables, \
	get_value_from_variables, test_date
from .variables import PipelineVariables


def get_topic_service(principal_service: PrincipalService) -> TopicService:
	return TopicService(principal_service)


class PossibleParameterType(str, Enum):
	STRING = 'string',
	NUMBER = 'number',
	BOOLEAN = 'boolean',
	DATE = 'date',
	TIME = 'time',
	DATETIME = 'datetime',
	ANY_SINGLE_VALUE = 'any-single-value',
	ANY_VALUE = 'any-value'


class ParsedStorageParameter:
	def __init__(
			self, parameter: Parameter, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool):
		self.parameter = parameter
		self.parse(parameter, available_schemas, principal_service, allow_in_memory_variables)

	# noinspection PyMethodMayBeStatic
	def find_topic(
			self, topic_id: Optional[TopicId], available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> Tuple[TopicSchema, bool]:
		"""
		find topic even it is not found in available list when in-memory variables is allowed.
		for the secondary one of return tuple, true when it is found in given available list,
		otherwise is false, which means it can be find in definitions
		"""
		if is_blank(topic_id):
			raise DataKernelException(f'Topic not declared.')
		schema = ArrayHelper(available_schemas).find(lambda x: x.get_topic().topicId == topic_id)
		if schema is None:
			if not allow_in_memory_variables:
				raise DataKernelException(f'Topic[id={topic_id}] not found.')
			else:
				# in pipeline, it might be from trigger data
				topic_service = get_topic_service(principal_service)
				topic: Optional[Topic] = topic_service.find_by_id(topic_id)
				if topic is None:
					raise DataKernelException(f'Topic[id={topic_id}] not found.')

				schema: Optional[TopicSchema] = topic_service.find_schema_by_name(
					topic.name, principal_service.get_tenant_id())
				if schema is None:
					raise DataKernelException(f'Topic schema[id={topic_id}] not found.')

				return schema, False
		else:
			return schema, True

	# noinspection PyMethodMayBeStatic
	def find_factor(self, factor_id: Optional[FactorId], topic: Topic) -> Factor:
		if is_blank(factor_id):
			raise DataKernelException(f'Factor not declared.')
		factor: Optional[Factor] = ArrayHelper(topic.factors).find(lambda x: x.factorId == factor_id)
		if factor is None:
			raise DataKernelException(
				f'Factor[id={factor_id}] in topic[id={topic.topicId}, name={topic.name}] not found.')
		return factor

	@abstractmethod
	def get_possible_types(self) -> List[PossibleParameterType]:
		pass

	@abstractmethod
	def parse(
			self, parameter: Parameter, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> None:
		pass

	@abstractmethod
	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		pass


def create_ask_factor_statement(
		schema: TopicSchema, factor: Factor) -> Callable[[PipelineVariables, PrincipalService], ColumnNameLiteral]:
	data_entity_helper = ask_topic_data_entity_helper(schema)

	return lambda variables, principal_service: ColumnNameLiteral(
		synonym=(schema.get_topic().kind == TopicKind.SYNONYM),
		entityName=schema.get_topic().name,
		columnName=data_entity_helper.get_column_name(factor.name)
	)


class ParsedStorageTopicFactorParameter(ParsedStorageParameter):
	# class variables, only for declare types, don't use it
	topic: Topic = None
	topicFromVariables: bool = False
	factor: Factor = None
	# in-memory value or a literal
	askValue: Union[
		Callable[[PipelineVariables, PrincipalService], Any],
		Callable[[PipelineVariables, PrincipalService], ColumnNameLiteral]
	] = None

	def get_possible_types(self) -> List[PossibleParameterType]:
		factor_type = self.factor.type
		if factor_type == FactorType.SEQUENCE:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.NUMBER:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.UNSIGNED:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.TEXT:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.ADDRESS:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.CONTINENT:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.REGION:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.COUNTRY:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.PROVINCE:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.CITY:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.DISTRICT:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.ROAD:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.COMMUNITY:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.FLOOR:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.RESIDENCE_TYPE:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.RESIDENTIAL_AREA:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.EMAIL:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.PHONE:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.MOBILE:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.FAX:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.DATETIME:
			return [PossibleParameterType.DATETIME]
		elif factor_type == FactorType.FULL_DATETIME:
			return [PossibleParameterType.DATETIME]
		elif factor_type == FactorType.DATE:
			return [PossibleParameterType.DATE]
		elif factor_type == FactorType.TIME:
			return [PossibleParameterType.TIME]
		elif factor_type == FactorType.YEAR:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.HALF_YEAR:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.QUARTER:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.MONTH:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.HALF_MONTH:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.TEN_DAYS:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.WEEK_OF_YEAR:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.WEEK_OF_MONTH:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.HALF_WEEK:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.DAY_OF_MONTH:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.DAY_OF_WEEK:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.DAY_KIND:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.HOUR:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.HOUR_KIND:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.MINUTE:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.SECOND:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.MILLISECOND:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.AM_PM:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.GENDER:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.OCCUPATION:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.DATE_OF_BIRTH:
			return [PossibleParameterType.DATETIME]
		elif factor_type == FactorType.AGE:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.ID_NO:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.RELIGION:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.NATIONALITY:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.BIZ_TRADE:
			return [PossibleParameterType.STRING]
		elif factor_type == FactorType.BIZ_SCALE:
			return [PossibleParameterType.NUMBER]
		elif factor_type == FactorType.BOOLEAN:
			return [PossibleParameterType.BOOLEAN]
		elif factor_type == FactorType.ENUM:
			return [PossibleParameterType.STRING]
		else:
			raise DataKernelException(f'Factor type[{factor_type}] is not supported.')

	def parse(
			self, parameter: TopicFactorParameter, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> None:
		# found in given schemas, means it is not from variables
		schema, found_in_given_available_schemas = self.find_topic(
			parameter.topicId, available_schemas, principal_service, allow_in_memory_variables)
		topic = schema.get_topic()
		self.topic = topic
		self.topicFromVariables = not found_in_given_available_schemas
		factor = self.find_factor(parameter.factorId, topic)
		self.factor = factor
		if found_in_given_available_schemas and is_raw_topic(topic) and not factor.flatten:
			raise DataKernelException(
				f'Factor[id={factor.factorId}, name={factor.name}] '
				f'on topic[id={topic.topicId}, name={topic.name}] is not flatten, '
				f'cannot be used on storage directly.')

		if not found_in_given_available_schemas:
			# get value from memory
			self.askValue = create_ask_factor_value(topic, factor)
		else:
			self.askValue = create_ask_factor_statement(schema, factor)

	def run(
			self,
			variables: PipelineVariables, principal_service: PrincipalService) -> Union[Any, ColumnNameLiteral]:
		return self.askValue(variables, principal_service)


def parse_variable_to_value(
		name: str, variables: PipelineVariables,
		available_schemas: List[TopicSchema], allow_in_memory_variables: bool,
		parse_from_variables: Callable[[PipelineVariables, PrincipalService, str], Tuple[bool, Any, Any]],
		principal_service: PrincipalService
) -> Tuple[bool, Union[Any, ParsedStorageParameter]]:
	"""
	the first element of the tuple is whether the variable is found in variables.
	if it is not from variables, the second element should be a topic factor parameter.
	"""
	if name.startswith('&'):
		# not from variables
		available_name = name[1:]
	elif allow_in_memory_variables:
		# try to get from memory variables
		parsed, value, parsed_value = parse_from_variables(variables, principal_service, name)
		if not parsed:
			raise DataKernelException(f'Value[{value}] cannot be parsed to date or datetime.')
		return True, parsed_value
	else:
		# still not from variables
		available_name = name

	if allow_in_memory_variables:
		# in pipeline "find by" use factor name. factor must find in given available schemas.
		# actually, the only one is the source topic of find by itself
		if len(available_schemas) == 0:
			raise DataKernelException(
				f'Variable name[{name}] is not supported, since no available topic given.')
		topic = available_schemas[0].get_topic()
		factor_name = available_name
	else:
		# in console subject use topic.factor. topic must in given available schemas
		if '.' not in available_name:
			raise DataKernelException(f'Variable name[{name}] is not supported.')
		names = available_name.split('.')
		if len(names) != 2:
			raise DataKernelException(f'Variable name[{name}] is not supported.')
		topic_name = names[0]
		factor_name = names[1]
		topic = ArrayHelper(available_schemas).map(lambda x: x.get_topic()).find(
			lambda x: x.name == topic_name)
		if topic is None:
			raise DataKernelException(f'Topic[{topic_name}] not found in given available topics.')
	factor: Optional[Factor] = ArrayHelper(available_schemas[0].get_topic().factors) \
		.find(lambda x: x.name == factor_name)
	if factor is None:
		raise DataKernelException(
			f'Factor[{factor_name}] in topic[id={topic.topicId}, name={topic.name}] not found.')
	return False, ParsedStorageTopicFactorParameter(
		TopicFactorParameter(kind=ParameterKind.TOPIC, topicId=topic.topicId, factorId=factor.factorId),
		available_schemas, principal_service, allow_in_memory_variables)


# noinspection DuplicatedCode
def create_date_diff(
		prefix: str, variable_name: str, function: VariablePredefineFunctions,
		available_schemas: List[TopicSchema], allow_in_memory_variables: bool
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	parsed_params = parse_function_in_variable(variable_name, function.value, 2)
	end_variable_name = parsed_params[0]
	start_variable_name = parsed_params[1]
	end_parsed, end_date = test_date(end_variable_name)
	start_parsed, start_date = test_date(start_variable_name)
	if end_parsed and start_parsed:
		# noinspection PyUnusedLocal,DuplicatedCode
		def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
			diff = compute_date_diff(function, end_date, start_date, variable_name)
			return diff if len(prefix) == 0 else f'{prefix}{diff}'

		return action
	else:
		def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
			if end_parsed:
				e_parsed = True
				e_date = end_date
			else:
				e_parsed, e_date = parse_variable_to_value(
					end_variable_name, variables, available_schemas, allow_in_memory_variables,
					get_date_from_variables, principal_service)
			if start_parsed:
				s_parsed = True,
				s_date = start_date
			else:
				s_parsed, s_date = parse_variable_to_value(
					start_variable_name, variables, available_schemas, allow_in_memory_variables,
					get_date_from_variables, principal_service)
			if e_parsed and s_parsed:
				diff = compute_date_diff(function, e_date, s_date, variable_name)
				return diff if len(prefix) == 0 else f'{prefix}{diff}'
			else:
				if function == VariablePredefineFunctions.YEAR_DIFF:
					operator = ComputedLiteralOperator.YEAR_DIFF
				elif function == VariablePredefineFunctions.MONTH_DIFF:
					operator = ComputedLiteralOperator.MONTH_DIFF
				elif function == VariablePredefineFunctions.DAY_DIFF:
					operator = ComputedLiteralOperator.DAY_DIFF
				else:
					raise DataKernelException(f'Variable name[{variable_name}] is not supported.')
				func = create_ask_value_for_computed(operator, [e_date, s_date])
				return func(variables, principal_service)

		return action


def create_date_format(
		prefix: str, variable_name: str,
		available_schemas: List[TopicSchema], allow_in_memory_variables: bool
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	parsed_params = parse_function_in_variable(variable_name, VariablePredefineFunctions.DATE_FORMAT.value, 2)
	variable_name = parsed_params[0]
	date_format = parsed_params[1]
	if is_blank(date_format):
		raise DataKernelException(f'Date format[{date_format}] cannot be recognized.')
	parsed, parsed_date = test_date(variable_name)
	if parsed:
		# noinspection PyUnusedLocal,DuplicatedCode
		def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
			translated_date_format = translate_date_format_to_memory(date_format)
			formatted = parsed_date.strftime(translated_date_format)
			return formatted if len(prefix) == 0 else f'{prefix}{formatted}'

		return action
	else:
		def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
			if parsed:
				date_parsed = True
				a_date = parsed_date
			else:
				date_parsed, a_date = parse_variable_to_value(
					variable_name, variables, available_schemas, allow_in_memory_variables,
					get_date_from_variables, principal_service)
			if date_parsed:
				translated_date_format = translate_date_format_to_memory(date_format)
				formatted = parsed_date.strftime(translated_date_format)
				return formatted if len(prefix) == 0 else f'{prefix}{formatted}'
			else:
				func = create_ask_value_for_computed(ComputedLiteralOperator.FORMAT_DATE, [a_date, date_format])
				return func(variables, principal_service)

		return action


def create_char_length(
		prefix: str, variable_name: str,
		available_schemas: List[TopicSchema], allow_in_memory_variables: bool
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		value_parsed, value = parse_variable_to_value(
			variable_name, variables, available_schemas, allow_in_memory_variables,
			get_value_from_variables, principal_service)
		if value_parsed:
			if value is None:
				length = 0
			elif isinstance(value, str):
				length = len(value)
			else:
				length = len(str(value))
			return length if len(prefix) == 0 else f'{prefix}{length}'
		else:
			func = create_ask_value_for_computed(ComputedLiteralOperator.CHAR_LENGTH, [value])
			return func(variables, principal_service)

	return action


# noinspection DuplicatedCode
def create_run_constant_segment(
		variable: MightAVariable, available_schemas: List[TopicSchema], allow_in_memory_variables: bool
) -> Tuple[Callable[[PipelineVariables, PrincipalService], Any], List[PossibleParameterType]]:
	prefix = variable.text
	has_prefix = len(prefix) != 0
	variable_name = variable.variable
	if variable_name == VariablePredefineFunctions.NEXT_SEQ.value:
		return \
			create_snowflake_generator(prefix), \
			[PossibleParameterType.STRING if has_prefix else PossibleParameterType.NUMBER]
	elif variable_name == VariablePredefineFunctions.NOW.value:
		if has_prefix:
			value = f'{prefix}{get_current_time_in_seconds().strftime("%Y-%m-%d %H:%M:%S")}'
			return lambda variables, principal_service: value, [PossibleParameterType.STRING]
		else:
			return lambda variables, principal_service: get_current_time_in_seconds(), [PossibleParameterType.DATETIME]
	elif variable_name.startswith(VariablePredefineFunctions.YEAR_DIFF.value):
		return \
			create_date_diff(
				prefix, variable_name,
				VariablePredefineFunctions.YEAR_DIFF, available_schemas, allow_in_memory_variables), \
			[PossibleParameterType.STRING if has_prefix else PossibleParameterType.NUMBER]
	elif variable_name.startswith(VariablePredefineFunctions.MONTH_DIFF.value):
		return \
			create_date_diff(
				prefix, variable_name,
				VariablePredefineFunctions.MONTH_DIFF, available_schemas, allow_in_memory_variables), \
			[PossibleParameterType.STRING if has_prefix else PossibleParameterType.NUMBER]
	elif variable_name.startswith(VariablePredefineFunctions.DAY_DIFF.value):
		return \
			create_date_diff(
				prefix, variable_name,
				VariablePredefineFunctions.DAY_DIFF, available_schemas, allow_in_memory_variables), \
			[PossibleParameterType.STRING if has_prefix else PossibleParameterType.NUMBER]
	elif variable_name.startswith(VariablePredefineFunctions.DATE_FORMAT.value):
		return \
			create_date_format(prefix, variable_name, available_schemas, allow_in_memory_variables), \
			[PossibleParameterType.STRING]
	elif variable_name.endswith(VariablePredefineFunctions.LENGTH.value):
		return \
			create_char_length(prefix, variable_name, available_schemas, allow_in_memory_variables), \
			[PossibleParameterType.STRING if has_prefix else PossibleParameterType.NUMBER]

	if allow_in_memory_variables:
		if variable_name.startswith(VariablePredefineFunctions.FROM_PREVIOUS_TRIGGER_DATA.value):
			if variable_name == VariablePredefineFunctions.FROM_PREVIOUS_TRIGGER_DATA.value:
				raise DataKernelException(
					f'Previous trigger data is a dict, cannot be used for storage. '
					f'Current constant segment is [{prefix}{{{variable_name}}}].')
			length = len(VariablePredefineFunctions.FROM_PREVIOUS_TRIGGER_DATA.value)
			if len(variable_name) < length + 2 or variable_name[length:length + 1] != '.':
				raise DataKernelException(f'Constant[{variable_name}] is not supported.')
			return \
				create_from_previous_trigger_data(prefix, variable_name[length + 1:]), \
				[PossibleParameterType.STRING if has_prefix else PossibleParameterType.ANY_VALUE]
		else:
			return \
				create_get_from_variables_with_prefix(prefix, variable_name), \
				[PossibleParameterType.STRING if has_prefix else PossibleParameterType.ANY_VALUE]
	else:
		# recover to original string
		return create_static_str(f'{prefix}{{{variable_name}}}'), [PossibleParameterType.STRING]


def create_ask_constant_value(
		variables: List[MightAVariable], available_schemas: List[TopicSchema], allow_in_memory_variables: bool
) -> Tuple[Callable[[PipelineVariables, PrincipalService], Any], List[PossibleParameterType]]:
	if len(variables) == 1:
		if variables[0].has_variable():
			return create_run_constant_segment(variables[0], available_schemas, allow_in_memory_variables)
		else:
			return create_static_str(variables[0].text), [PossibleParameterType.STRING]
	else:
		segments = ArrayHelper(variables) \
			.map(lambda x: create_run_constant_segment(x, available_schemas, allow_in_memory_variables)) \
			.map(lambda x: x[0]) \
			.to_list()
		return create_ask_value_for_computed(ComputedLiteralOperator.CONCAT, segments), [PossibleParameterType.STRING]


class ParsedStorageConstantParameter(ParsedStorageParameter):
	# class variables, only for declare types, don't use it
	askValue: Callable[[PipelineVariables, PrincipalService], Any] = None
	possibleTypes: List[PossibleParameterType] = []

	def get_possible_types(self) -> List[PossibleParameterType]:
		return self.possibleTypes

	# noinspection DuplicatedCode
	def parse(
			self, parameter: ConstantParameter, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> None:
		value = parameter.value
		if value is None:
			self.possibleTypes = [PossibleParameterType.ANY_SINGLE_VALUE]
			self.askValue = always_none
		elif len(value) == 0:
			self.possibleTypes = [PossibleParameterType.ANY_SINGLE_VALUE]
			self.askValue = always_none
		elif is_blank(value):
			self.possibleTypes = [PossibleParameterType.ANY_SINGLE_VALUE]
			self.askValue = always_none
		elif '{' not in value or '}' not in value:
			self.possibleTypes = [PossibleParameterType.ANY_VALUE]
			self.askValue = create_static_str(value)
		else:
			_, variables = parse_variable(value)
			self.askValue, self.possibleTypes = create_ask_constant_value(
				variables, available_schemas, allow_in_memory_variables)

	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		return self.askValue(variables, principal_service)


class TypedParsedStorageConstantParameter(ParsedStorageConstantParameter):
	def __init__(
			self,
			parameter: ParsedStorageConstantParameter,
			available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool,
			try_to_type: Callable[[Any], Any]):
		# simply pass to super constructor
		super().__init__(parameter.parameter, available_schemas, principal_service, allow_in_memory_variables)
		self.parsedParameter = parameter
		if try_to_type is None:
			raise DataKernelException('Type cast function cannot be none.')
		self.tryToType = try_to_type

	def parse(
			self,
			parameter: ConstantParameter, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> None:
		# ignore
		pass

	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		value = self.parsedParameter.run(variables, principal_service)
		return self.tryToType(value)


def create_ask_value_for_computed(
		operator: ComputedLiteralOperator,
		elements: List[Union[
			ParsedStorageParameter,
			Tuple[ParsedStorageCondition, ParsedStorageParameter],
			Callable[[PipelineVariables, PrincipalService], Any],
			Any]]
) -> Callable[[PipelineVariables, PrincipalService], Any]:
	def compute(variables: PipelineVariables, principal_service: PrincipalService) -> Literal:
		def transform(
				element: Union[ParsedStorageParameter, Tuple[ParsedStorageCondition, ParsedStorageParameter], Any]
		) -> Union[Literal, Union[EntityCriteriaStatement, Literal]]:
			if isinstance(element, ParsedStorageParameter):
				return element.run(variables, principal_service)
			elif isinstance(element, Tuple):
				condition: ParsedStorageCondition = element[0]
				parameter: ParsedStorageParameter = element[1]
				return condition.run(variables, principal_service), parameter.run(variables, principal_service)
			elif isfunction(element):
				return element(variables, principal_service)
			else:
				return element

		return ComputedLiteral(operator=operator, elements=ArrayHelper(elements).map(transform).to_list())

	return compute


class ParsedStorageComputedParameter(ParsedStorageParameter):
	# class variables, only for declare types, don't use it
	askValue: Callable[[PipelineVariables, PrincipalService], Literal] = None
	possibleTypes: List[PossibleParameterType] = []

	def get_possible_types(self) -> List[PossibleParameterType]:
		return self.possibleTypes

	def parse(
			self, parameter: ComputedParameter, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> None:
		compute_type = parameter.type
		if is_blank(compute_type) or compute_type == ParameterComputeType.NONE:
			raise DataKernelException(f'Compute type not declared.')

		def parse_parameter(param: Parameter) -> ParsedStorageParameter:
			return parse_parameter_for_storage(
				param, available_schemas, principal_service, allow_in_memory_variables)

		def try_to_date_type(value: Any) -> Any:
			if value is None:
				return None
			if isinstance(value, date):
				return value
			if isinstance(value, str):
				parsed, dt_value = is_date(value, ask_all_date_formats())
				return dt_value if parsed else value
			else:
				return value

		def parse_typed_parameter(param: Parameter, try_to_type: Callable[[Any], Any]) -> ParsedStorageParameter:
			parsed = parse_parameter_for_storage(
				param, available_schemas, principal_service, allow_in_memory_variables)
			if isinstance(parsed, ParsedStorageConstantParameter):
				return TypedParsedStorageConstantParameter(
					parsed, available_schemas, principal_service, allow_in_memory_variables, try_to_type)
			else:
				return parsed

		def parse_sub_parameters(param: ComputedParameter) -> List[ParsedStorageParameter]:
			return ArrayHelper(param.parameters).map(parse_parameter).to_list()

		self.possibleTypes = [PossibleParameterType.NUMBER]
		if compute_type == ParameterComputeType.ADD:
			assert_parameter_count('add', parameter.parameters, 2)
			# treat none value as 0
			self.askValue = create_ask_value_for_computed(ComputedLiteralOperator.ADD, parse_sub_parameters(parameter))
		elif compute_type == ParameterComputeType.SUBTRACT:
			assert_parameter_count('subtract', parameter.parameters, 2)
			# treat none value as 0
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.SUBTRACT, parse_sub_parameters(parameter))
		elif compute_type == ParameterComputeType.MULTIPLY:
			assert_parameter_count('multiply', parameter.parameters, 2)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.MULTIPLY, parse_sub_parameters(parameter))
		elif compute_type == ParameterComputeType.DIVIDE:
			assert_parameter_count('divide', parameter.parameters, 2)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.DIVIDE, parse_sub_parameters(parameter))
		elif compute_type == ParameterComputeType.MODULUS:
			assert_parameter_count('modulus', parameter.parameters, 2)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.MODULUS, parse_sub_parameters(parameter))
		elif compute_type == ParameterComputeType.YEAR_OF:
			assert_parameter_count('year-of', parameter.parameters, 1, 1)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.YEAR_OF, [parse_typed_parameter(parameter.parameters[0], try_to_date_type)])
		elif compute_type == ParameterComputeType.HALF_YEAR_OF:
			assert_parameter_count('half-year-of', parameter.parameters, 1, 1)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.HALF_YEAR_OF,
				[parse_typed_parameter(parameter.parameters[0], try_to_date_type)])
		elif compute_type == ParameterComputeType.QUARTER_OF:
			assert_parameter_count('quarter-of', parameter.parameters, 1, 1)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.QUARTER_OF, [parse_typed_parameter(parameter.parameters[0], try_to_date_type)])
		elif compute_type == ParameterComputeType.MONTH_OF:
			assert_parameter_count('month-of', parameter.parameters, 1, 1)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.MONTH_OF, [parse_typed_parameter(parameter.parameters[0], try_to_date_type)])
		elif compute_type == ParameterComputeType.WEEK_OF_YEAR:
			assert_parameter_count('week-of-year', parameter.parameters, 1, 1)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.WEEK_OF_YEAR,
				[parse_typed_parameter(parameter.parameters[0], try_to_date_type)])
		elif compute_type == ParameterComputeType.WEEK_OF_MONTH:
			assert_parameter_count('week-of-month', parameter.parameters, 1, 1)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.WEEK_OF_MONTH,
				[parse_typed_parameter(parameter.parameters[0], try_to_date_type)])
		elif compute_type == ParameterComputeType.DAY_OF_MONTH:
			assert_parameter_count('day-of-month', parameter.parameters, 1, 1)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.DAY_OF_MONTH,
				[parse_typed_parameter(parameter.parameters[0], try_to_date_type)])
		elif compute_type == ParameterComputeType.DAY_OF_WEEK:
			assert_parameter_count('day-of-week', parameter.parameters, 1, 1)
			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.DAY_OF_WEEK, [parse_typed_parameter(parameter.parameters[0], try_to_date_type)])
		elif compute_type == ParameterComputeType.CASE_THEN:
			# noinspection DuplicatedCode
			assert_parameter_count('case-then', parameter.parameters, 1)
			cases = parameter.parameters
			if cases is None or len(cases) == 0:
				raise DataKernelException(f'Case not declared in case then computation.')

			anyways = ArrayHelper(cases).filter(lambda x: not x.conditional).to_list()
			if len(anyways) > 1:
				raise DataKernelException(
					f'Multiple anyway routes declared in case then computation[{parameter.dict()}].')

			def parse_route(param: Parameter) -> Tuple[ParsedStorageCondition, ParsedStorageParameter]:
				return parse_conditional_parameter_for_storage(
					param, available_schemas, principal_service, allow_in_memory_variables)

			# noinspection DuplicatedCode
			routes = ArrayHelper(cases).filter(lambda x: x.conditional).map(parse_route).to_list()
			anyway = anyways[0] if len(anyways) == 1 else None
			if anyway is not None:
				anyway_route = parse_parameter(anyway)
			else:
				anyway_route = None

			self.possibleTypes = ArrayHelper(routes) \
				.map(lambda x: x[1].get_possible_types()) \
				.flatten() \
				.grab(*([] if anyway_route is None else anyway_route.get_possible_types())) \
				.distinct() \
				.to_list()

			self.askValue = create_ask_value_for_computed(
				ComputedLiteralOperator.CASE_THEN, ArrayHelper(routes).grab(anyway_route).to_list())
		else:
			raise DataKernelException(f'Compute type[{compute_type}] is not supported.')

	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> ComputedLiteral:
		return self.askValue(variables, principal_service)


def parse_parameter_for_storage(
		parameter: Optional[Parameter], available_schemas: List[TopicSchema],
		principal_service: PrincipalService, allow_in_memory_variables: bool
) -> ParsedStorageParameter:
	if parameter is None:
		raise DataKernelException('Parameter cannot be none.')
	elif isinstance(parameter, TopicFactorParameter):
		return ParsedStorageTopicFactorParameter(
			parameter, available_schemas, principal_service, allow_in_memory_variables)
	elif isinstance(parameter, ConstantParameter):
		return ParsedStorageConstantParameter(
			parameter, available_schemas, principal_service, allow_in_memory_variables)
	elif isinstance(parameter, ComputedParameter):
		return ParsedStorageComputedParameter(
			parameter, available_schemas, principal_service, allow_in_memory_variables)
	else:
		raise DataKernelException(f'Parameter[{parameter.dict()}] is not supported.')


class ParsedStorageCondition:
	def __init__(
			self, condition: ParameterCondition, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool):
		self.condition = condition
		self.parse(condition, available_schemas, principal_service, allow_in_memory_variables)

	@abstractmethod
	def parse(
			self, condition: ParameterCondition, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> None:
		pass

	@abstractmethod
	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> EntityCriteriaStatement:
		pass


class ParsedStorageJoint(ParsedStorageCondition):
	# class variables, only for declare types, don't use it
	jointType: ParameterJointType = ParameterJointType.AND
	filters: List[ParsedStorageCondition] = []

	def parse(
			self, condition: ParameterJoint, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> None:
		self.jointType = ParameterJointType.OR \
			if condition.jointType == ParameterJointType.OR else ParameterJointType.AND

		def parse_sub(sub: ParameterCondition) -> ParsedStorageCondition:
			return parse_condition_for_storage(sub, available_schemas, principal_service, allow_in_memory_variables)

		self.filters = ArrayHelper(condition.filters).map(parse_sub).to_list()

	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> EntityCriteriaJoint:
		if self.jointType == ParameterJointType.OR:
			return EntityCriteriaJoint(
				conjunction=EntityCriteriaJointConjunction.OR,
				children=ArrayHelper(self.filters).map(lambda x: x.run(variables, principal_service)).to_list()
			)
		else:
			# and or not given
			return EntityCriteriaJoint(
				conjunction=EntityCriteriaJointConjunction.AND,
				children=ArrayHelper(self.filters).map(lambda x: x.run(variables, principal_service)).to_list()
			)


class ParsedStorageExpression(ParsedStorageCondition):
	# class variables, only for declare types, don't use it
	left: Optional[ParsedStorageParameter] = None
	operator: Optional[ParameterExpressionOperator] = None
	right: Optional[ParsedStorageParameter] = None

	def parse(
			self, condition: ParameterCondition, available_schemas: List[TopicSchema],
			principal_service: PrincipalService, allow_in_memory_variables: bool) -> None:
		self.left = parse_parameter_for_storage(
			condition.left, available_schemas, principal_service, allow_in_memory_variables)
		self.operator = condition.operator
		if self.operator != ParameterExpressionOperator.EMPTY and self.operator != ParameterExpressionOperator.NOT_EMPTY:
			self.right = parse_parameter_for_storage(
				condition.right, available_schemas, principal_service, allow_in_memory_variables)

	# noinspection PyMethodMayBeStatic,DuplicatedCode
	def handle_equation_possible_types(self, a_value: Any, another_possible_types: List[PossibleParameterType]) -> Any:
		if isinstance(a_value, (int, float, Decimal, str)):
			if another_possible_types == [PossibleParameterType.BOOLEAN]:
				if str(a_value).lower() in ['1', 't', 'true', 'y', 'yes']:
					return Decimal('1')
				elif str(a_value).lower() in ['0', 'f', 'false', 'n', 'no']:
					return Decimal('0')
			elif another_possible_types == [PossibleParameterType.NUMBER]:
				parsed, decimal_value = is_decimal(str(a_value))
				if parsed:
					return decimal_value
		return a_value

	# noinspection PyMethodMayBeStatic,DuplicatedCode
	def handle_comparison_possible_types(
			self, a_value: Any, another_possible_types: List[PossibleParameterType]) -> Any:
		if isinstance(a_value, str):
			if PossibleParameterType.NUMBER in another_possible_types:
				parsed, decimal_value = is_decimal(a_value)
				if parsed:
					return decimal_value
			if PossibleParameterType.DATE in another_possible_types or PossibleParameterType.DATETIME in another_possible_types:
				parsed, date_value = is_date(a_value, ask_all_date_formats())
				if parsed:
					return date_value
			if PossibleParameterType.TIME in another_possible_types:
				parsed, time_value = is_time(a_value, ask_time_formats())
				if parsed:
					return time_value
		return a_value

	def run(self, variables: PipelineVariables, principal_service: PrincipalService) -> EntityCriteriaExpression:
		left_value = self.left.run(variables, principal_service)
		if self.operator == ParameterExpressionOperator.EMPTY:
			return EntityCriteriaExpression(left=left_value, operator=EntityCriteriaOperator.IS_EMPTY)
		elif self.operator == ParameterExpressionOperator.NOT_EMPTY:
			return EntityCriteriaExpression(left=left_value, operator=EntityCriteriaOperator.IS_NOT_EMPTY)

		right_value = self.right.run(variables, principal_service)
		if self.operator == ParameterExpressionOperator.EQUALS:
			left_value = self.handle_equation_possible_types(left_value, self.right.get_possible_types())
			right_value = self.handle_equation_possible_types(right_value, self.left.get_possible_types())
			return EntityCriteriaExpression(
				left=left_value, operator=EntityCriteriaOperator.EQUALS, right=right_value)
		elif self.operator == ParameterExpressionOperator.NOT_EQUALS:
			left_value = self.handle_equation_possible_types(left_value, self.right.get_possible_types())
			right_value = self.handle_equation_possible_types(right_value, self.left.get_possible_types())
			return EntityCriteriaExpression(
				left=left_value, operator=EntityCriteriaOperator.NOT_EQUALS, right=right_value)
		elif self.operator == ParameterExpressionOperator.LESS:
			left_value = self.handle_comparison_possible_types(left_value, self.right.get_possible_types())
			right_value = self.handle_comparison_possible_types(right_value, self.left.get_possible_types())
			return EntityCriteriaExpression(
				left=left_value, operator=EntityCriteriaOperator.LESS_THAN, right=right_value)
		elif self.operator == ParameterExpressionOperator.LESS_EQUALS:
			left_value = self.handle_comparison_possible_types(left_value, self.right.get_possible_types())
			right_value = self.handle_comparison_possible_types(right_value, self.left.get_possible_types())
			return EntityCriteriaExpression(
				left=left_value, operator=EntityCriteriaOperator.LESS_THAN_OR_EQUALS, right=right_value)
		elif self.operator == ParameterExpressionOperator.MORE:
			left_value = self.handle_comparison_possible_types(left_value, self.right.get_possible_types())
			right_value = self.handle_comparison_possible_types(right_value, self.left.get_possible_types())
			return EntityCriteriaExpression(
				left=left_value, operator=EntityCriteriaOperator.GREATER_THAN, right=right_value)
		elif self.operator == ParameterExpressionOperator.MORE_EQUALS:
			left_value = self.handle_comparison_possible_types(left_value, self.right.get_possible_types())
			right_value = self.handle_comparison_possible_types(right_value, self.left.get_possible_types())
			return EntityCriteriaExpression(
				left=left_value, operator=EntityCriteriaOperator.GREATER_THAN_OR_EQUALS, right=right_value)
		elif self.operator == ParameterExpressionOperator.IN:
			return EntityCriteriaExpression(
				left=left_value, operator=EntityCriteriaOperator.IN, right=right_value)
		elif self.operator == ParameterExpressionOperator.NOT_IN:
			return EntityCriteriaExpression(
				left=left_value, operator=EntityCriteriaOperator.NOT_IN, right=right_value)
		else:
			raise DataKernelException(
				f'Operator[{self.operator}] is not supported, found from expression[{self.condition.dict()}].')


def parse_condition_for_storage(
		condition: Optional[ParameterCondition], available_schemas: List[TopicSchema],
		principal_service: PrincipalService, allow_in_memory_variables: bool) -> ParsedStorageCondition:
	if condition is None:
		raise DataKernelException('Condition cannot be none.')
	if isinstance(condition, ParameterJoint):
		return ParsedStorageJoint(condition, available_schemas, principal_service, allow_in_memory_variables)
	elif isinstance(condition, ParameterExpression):
		return ParsedStorageExpression(condition, available_schemas, principal_service, allow_in_memory_variables)
	else:
		raise DataKernelException(f'Condition[{condition.dict()}] is not supported.')


def parse_conditional_parameter_for_storage(
		parameter: Parameter, available_schemas: List[TopicSchema],
		principal_service: PrincipalService, allow_in_memory_variables: bool
) -> Tuple[ParsedStorageCondition, ParsedStorageParameter]:
	return \
		parse_condition_for_storage(parameter.on, available_schemas, principal_service, allow_in_memory_variables), \
		parse_parameter_for_storage(parameter, available_schemas, principal_service, allow_in_memory_variables)


class ParsedStorageMappingFactor:
	def __init__(
			self, schema: TopicSchema, mapping_factor: MappingFactor, accumulate_mode: Optional[AccumulateMode],
			principal_service: PrincipalService):
		self.mappingFactor = mapping_factor
		self.arithmetic = AggregateArithmetic.NONE if mapping_factor.arithmetic is None else mapping_factor.arithmetic
		self.accumulateMode = AccumulateMode.STANDARD if accumulate_mode is None else accumulate_mode
		factor_id = mapping_factor.factorId
		if is_blank(factor_id):
			raise DataKernelException('Factor not declared on mapping.')
		topic = schema.get_topic()
		factor: Optional[Factor] = ArrayHelper(topic.factors).find(lambda x: x.factorId == factor_id)
		if factor is None:
			raise DataKernelException(
				f'Factor[id={factor_id}] in topic[id={topic.topicId}, name={topic.name}] not found.')
		self.factor = factor
		if mapping_factor.source is None:
			raise DataKernelException('Source of mapping factor not declared.')
		# parameter in mapping factor always retrieve value from variables
		self.parsedParameter = parse_parameter_in_memory(mapping_factor.source, principal_service)

	# noinspection PyMethodMayBeStatic
	def get_aggregate_assist_column(self, data: Dict[str, Any], return_default: bool):
		value = data.get(TopicDataColumnNames.AGGREGATE_ASSIST.value)
		return {} if return_default and value is None else value

	# noinspection PyMethodMayBeStatic
	def set_aggregate_assist_column(self, data: Dict[str, Any], assist: Dict[str, Decimal]) -> None:
		data[TopicDataColumnNames.AGGREGATE_ASSIST.value] = assist

	def set_aggregate_assist_avg_count(self, data: Dict[str, Any], value: int) -> None:
		assist = self.get_aggregate_assist_column(data, True)
		assist[f'{self.factor.name}.avg_count'] = value
		self.set_aggregate_assist_column(data, assist)

	def get_aggregate_assist_avg_count(self, data: Dict[str, Any]) -> int:
		assist = self.get_aggregate_assist_column(data, True)
		value = assist.get(f'{self.factor.name}.avg_count')
		parsed, decimal_value = is_decimal(value)
		return decimal_value if parsed else 0

	def get_original_value(self, original_data: Optional[Dict[str, Any]]) -> int:
		"""
		return 0 when not found or cannot be parsed to decimal
		"""
		value = original_data.get(self.factor.name)
		parsed, decimal_value = is_decimal(value)
		if parsed:
			return int(decimal_value)
		else:
			return 0

	def compute_current_value(self, variables: PipelineVariables, principal_service: PrincipalService) -> Decimal:
		value = self.parsedParameter.value(variables, principal_service)
		parsed, decimal_value = is_decimal(value)
		decimal_value = 0 if not parsed else decimal_value
		return decimal_value

	def compute_previous_value(self, variables: PipelineVariables, principal_service: PrincipalService) -> Decimal:
		value = self.parsedParameter.value(variables.backward_to_previous(), principal_service)
		parsed, decimal_value = is_decimal(value)
		decimal_value = 0 if not parsed else decimal_value
		return decimal_value

	def compute_previous_and_current_value(
			self, variables: PipelineVariables, principal_service: PrincipalService) -> Tuple[Decimal, Decimal]:
		return \
			self.compute_previous_value(variables, principal_service), \
			self.compute_current_value(variables, principal_service)

	def run(
			self, data: Dict[str, Any], original_data: Optional[Dict[str, Any]], variables: PipelineVariables,
			principal_service: PrincipalService
	) -> Tuple[str, Any]:
		if self.arithmetic == AggregateArithmetic.SUM:
			if original_data is None:
				# the very first time to insert this
				if self.accumulateMode == AccumulateMode.REVERSE:
					raise DataKernelException(
						'There is no existing aggregated data, therefore reverse accumulating cannot be performed here.')
				else:
					# simply set as value
					value = self.compute_current_value(variables, principal_service)
			elif self.accumulateMode == AccumulateMode.CUMULATE:
				# using force cumulating explicitly, ignore previous data anyway
				value = \
					self.compute_current_value(variables, principal_service) + self.get_original_value(original_data)
			elif variables.has_previous_trigger_data():
				# has previous data, it used to be triggered,
				if self.accumulateMode == AccumulateMode.REVERSE:
					# reverse the last accumulating
					value = \
						self.get_original_value(original_data) \
						- self.compute_previous_value(variables, principal_service)
				else:
					# standard mode, then subtract previous value and add current value
					previous_value, current_value = \
						self.compute_previous_and_current_value(variables, principal_service)
					value = self.get_original_value(original_data) - previous_value + current_value
			elif self.accumulateMode == AccumulateMode.REVERSE:
				# no previous data, reverse cannot be performed
				raise DataKernelException(
					'There is no previous trigger data, therefore reverse accumulating cannot be performed here.')
			else:
				# no previous data, then find original value, add current value
				value = \
					self.compute_current_value(variables, principal_service) + self.get_original_value(original_data)
		elif self.arithmetic == AggregateArithmetic.AVG:
			# noinspection DuplicatedCode
			if original_data is None:
				# the very first time to insert this
				if self.accumulateMode == AccumulateMode.REVERSE:
					raise DataKernelException(
						'There is no existing aggregated data, therefore reverse accumulating cannot be performed here.')
				else:
					# simply set as value
					value = self.parsedParameter.value(variables, principal_service)
					self.set_aggregate_assist_avg_count(data, 1)
			elif self.accumulateMode == AccumulateMode.CUMULATE:
				# using force cumulating explicitly, ignore previous data anyway
				current_value = self.compute_current_value(variables, principal_service)
				count = self.get_aggregate_assist_avg_count(original_data)
				original_value = self.get_original_value(original_data)
				value = (original_value * count + current_value) / (count + 1)
				self.set_aggregate_assist_avg_count(data, count + 1)
			elif variables.has_previous_trigger_data():
				# has previous data, it used to be triggered
				count = self.get_aggregate_assist_avg_count(original_data)
				# should at least has one item, 0 must be an error value, fix it
				if count == 0:
					raise DataKernelException(
						'No existing accumulated average (count is 0), '
						'therefore average accumulating on data change cannot be performed here.')
				elif self.accumulateMode == AccumulateMode.REVERSE:
					# reverse the last accumulating
					if count == 1:
						# the last one
						value = 0
						self.set_aggregate_assist_avg_count(data, 0)
					else:
						original_value = self.get_original_value(original_data)
						previous_value = self.compute_previous_value(variables, principal_service)
						value = (original_value * count - previous_value) / (count - 1)
						self.set_aggregate_assist_avg_count(data, count - 1)
				else:
					# standard mode
					# find previous value and avg count in original data, to compute the new avg
					previous_value, current_value = \
						self.compute_previous_and_current_value(variables, principal_service)
					original_value = self.get_original_value(original_data)
					value = (original_value * count + current_value - previous_value) / count
					self.set_aggregate_assist_avg_count(data, count)
			elif self.accumulateMode == AccumulateMode.REVERSE:
				# no previous data, reverse cannot be performed
				raise DataKernelException(
					'There is no previous trigger data, therefore reverse accumulating cannot be performed here.')
			else:
				# no previous data, data is triggered at first time,
				# find original value, add current value, to compute the new avg
				current_value = self.compute_current_value(variables, principal_service)
				count = self.get_aggregate_assist_avg_count(original_data)
				original_value = self.get_original_value(original_data)
				value = (original_value * count + current_value) / (count + 1)
				self.set_aggregate_assist_avg_count(data, count + 1)
		elif self.arithmetic == AggregateArithmetic.COUNT:
			if original_data is None:
				# the very first time to insert this
				if self.accumulateMode == AccumulateMode.REVERSE:
					raise DataKernelException(
						'There is no existing aggregated data, therefore reverse count cannot be performed here.')
				else:
					# count always be 1
					value = 1
			elif self.accumulateMode == AccumulateMode.CUMULATE:
				# using force cumulating explicitly, ignore previous data anyway, always add 1
				value = self.get_original_value(original_data) + 1
			elif variables.has_previous_trigger_data():
				# has previous data, it used to be triggered
				value = self.get_original_value(original_data)
				if self.accumulateMode == AccumulateMode.REVERSE:
					# reverse the last accumulating, subtract 1
					if value <= 0:
						raise DataKernelException('Count <= 0, therefore reverse count cannot be performed here.')
					else:
						value = value - 1
				else:
					# standard mode, since already be counted before, ignore
					# but when there is some incorrect value already saved, correct it to 1
					value = 1 if value == 0 else value
			elif self.accumulateMode == AccumulateMode.REVERSE:
				# no previous data, reverse cannot be performed
				raise DataKernelException(
					'There is no previous trigger data, therefore reverse count cannot be performed here.')
			else:
				# no previous data, count + 1
				value = self.get_original_value(original_data) + 1
		else:
			value = self.parsedParameter.value(variables, principal_service)
			# value for mapping is computed in memory
			# parse factor value to applicable type
			value = cast_value_for_factor(value, self.factor) if value is not None else None
		return self.factor.name, value


class ParsedStorageMapping:
	def __init__(
			self, schema: TopicSchema, mapping: Optional[List[MappingFactor]],
			accumulate_mode: Optional[AccumulateMode],
			principal_service: PrincipalService):
		self.mapping = mapping
		if mapping is None:
			raise DataKernelException('Mapping cannot be none.')
		if len(mapping) == 0:
			raise DataKernelException('At least one mapping is required.')
		self.parsedMappingFactors = ArrayHelper(mapping).map(
			lambda x: ParsedStorageMappingFactor(schema, x, accumulate_mode, principal_service)).to_list()

	def run(
			self,
			original_data: Optional[Dict[str, Any]], variables: PipelineVariables,
			principal_service: PrincipalService
	) -> Dict[str, Any]:
		return ArrayHelper(self.parsedMappingFactors).reduce(
			lambda data, x: self.run_factor(data, x, original_data, variables, principal_service), {})

	# noinspection PyMethodMayBeStatic
	def run_factor(
			self, data: Dict[str, Any], parsed_factor: ParsedStorageMappingFactor,
			original_data: Optional[Dict[str, Any]], variables: PipelineVariables,
			principal_service: PrincipalService) -> Dict[str, Any]:
		factor_name, value = parsed_factor.run(data, original_data, variables, principal_service)
		data[factor_name] = value
		return data


def parse_mapping_for_storage(
		schema: TopicSchema, action: Union[ToTopic, FromTopic],
		mapping: Optional[List[MappingFactor]], principal_service: PrincipalService) -> ParsedStorageMapping:
	if isinstance(action, InsertRowAction):
		# for insert action, always be standard mode
		accumulate_mode = AccumulateMode.STANDARD
	elif isinstance(action, InsertOrMergeRowAction):
		accumulate_mode = AccumulateMode.STANDARD if action.accumulateMode is None else action.accumulateMode
		# reverse is not allowed for insert/merge action
		accumulate_mode = AccumulateMode.STANDARD if accumulate_mode == AccumulateMode.REVERSE else accumulate_mode
	elif isinstance(action, WriteTopicAction):
		# merge row or write topic
		accumulate_mode = AccumulateMode.STANDARD if action.accumulateMode is None else action.accumulateMode
	else:
		# not required for other actions
		accumulate_mode = None

	return ParsedStorageMapping(schema, mapping, accumulate_mode, principal_service)



from abc import abstractmethod
from typing import List, Optional

from watchmen_model.admin import Pipeline
from watchmen_model.common import PipelineId
from watchmen_utilities import ArrayHelper
from .cache_manager import get_pipeline_by_id_cache
from .internal_cache import InternalCache
from .pipeline_by_topic_cache import pipeline_by_topic_cache


class PipelineCacheListener:
	@abstractmethod
	def on_pipeline_added(self, pipeline: Pipeline) -> None:
		pass

	@abstractmethod
	def on_pipeline_removed(self, pipeline: Pipeline) -> None:
		pass

	@abstractmethod
	def on_cache_cleared(self) -> None:
		pass


class PipelineCache:
	listeners: List[PipelineCacheListener] = []

	def __init__(self):
		self.byIdCache = InternalCache(cache=get_pipeline_by_id_cache)

	def add_cache_listener(self, listener: PipelineCacheListener) -> None:
		if listener not in self.listeners:
			self.listeners.append(listener)

	def fire_pipeline_added(self, pipeline: Pipeline):
		ArrayHelper(self.listeners).each(lambda x: x.on_pipeline_added(pipeline))

	def remove_cache_listener(self, listener: PipelineCacheListener) -> None:
		if listener in self.listeners:
			self.listeners.remove(listener)

	def fire_pipeline_removed(self, pipeline: Pipeline):
		ArrayHelper(self.listeners).each(lambda x: x.on_pipeline_removed(pipeline))

	def put(self, pipeline: Pipeline) -> Optional[Pipeline]:
		existing: Optional[Pipeline] = self.byIdCache.put(pipeline.pipelineId, pipeline)
		if existing is not None:
			self.fire_pipeline_removed(pipeline)
			if existing.topicId != pipeline.topicId:
				# trigger topic changed
				pipeline_by_topic_cache.remove_one(existing.topicId, existing.pipelineId)
				pipeline_by_topic_cache.append_one(pipeline.topicId, pipeline.pipelineId)
		else:
			# new pipline
			pipeline_by_topic_cache.append_one(pipeline.topicId, pipeline.pipelineId)
		self.fire_pipeline_added(pipeline)
		return existing

	def get(self, pipeline_id: PipelineId) -> Optional[Pipeline]:
		return self.byIdCache.get(pipeline_id)

	def remove(self, pipeline_id: PipelineId) -> Optional[Pipeline]:
		existing: Optional[Pipeline] = self.byIdCache.remove(pipeline_id)
		if existing is not None:
			pipeline_by_topic_cache.remove_one(existing.topicId, existing.pipelineId)
		return existing

	def all(self) -> List[Pipeline]:
		return list(self.byIdCache.values())

	def clear(self) -> None:
		self.byIdCache.clear()
		self.fire_cache_cleared()
		pipeline_by_topic_cache.clear()

	def fire_cache_cleared(self):
		ArrayHelper(self.listeners).each(lambda x: x.on_cache_cleared())


pipeline_cache = PipelineCache()



from typing import Callable, List, Optional

from watchmen_model.common import DataSourceId
from watchmen_model.system import DataSource
from watchmen_storage import TopicDataStorageSPI
from .cache_manager import get_data_source_by_id_cache, get_data_storage_builder_by_id_cache
from .internal_cache import InternalCache


class DataSourceCache:
	"""
	data source cache will not impact other caches
	"""

	def __init__(self):
		self.byIdCache = InternalCache(cache=get_data_source_by_id_cache)
		self.builderByIdCache = InternalCache(cache=get_data_storage_builder_by_id_cache)

	def put(self, data_source: DataSource) -> Optional[DataSource]:
		return self.byIdCache.put(data_source.dataSourceId, data_source)

	def put_builder(
			self, data_source_id: DataSourceId, builder: Callable[[], TopicDataStorageSPI]
	) -> Optional[Callable[[], TopicDataStorageSPI]]:
		return self.builderByIdCache.put(data_source_id, builder)

	def get(self, data_source_id: DataSourceId) -> Optional[DataSource]:
		return self.byIdCache.get(data_source_id)

	def get_builder(self, data_source_id: DataSourceId) -> Optional[Callable[[], TopicDataStorageSPI]]:
		return self.builderByIdCache.get(data_source_id)

	def remove(self, data_source_id: DataSourceId) -> Optional[DataSource]:
		data_source = self.byIdCache.remove(data_source_id)
		self.builderByIdCache.remove(data_source_id)
		return data_source

	def all(self) -> List[DataSource]:
		return list(self.byIdCache.values())

	def clear(self) -> None:
		self.byIdCache.clear()


data_source_cache = DataSourceCache()



from .aes_encryptor import AESEncryptor
from .center_masker import CenterMasker
from .date_masker import DateMasker
from .encryptor import Encryptor
from .encryptor_registry import find_encryptor, is_encryptor_registered, register_encryptor
from .last_masker import LastMasker
from .mail_masker import MailMasker
from .md5_encryptor import MD5Encryptor
from .sha256_encryptor import SHA256Encryptor



from abc import abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from watchmen_auth import PrincipalService
from watchmen_data_kernel.topic_schema import TopicSchema
from watchmen_model.admin import Topic
from watchmen_model.common import Pageable, TenantId
from watchmen_model.pipeline_kernel import TopicDataColumnNames
from watchmen_storage import ColumnNameLiteral, EntityColumnName, EntityCriteria, EntityCriteriaExpression, \
	EntityDeleter, EntityDistinctValuesFinder, EntityFinder, EntityHelper, EntityIdHelper, EntityPager, EntitySort, \
	EntityStraightColumn, EntityStraightValuesFinder, EntityUpdate, EntityUpdater, SnowflakeGenerator
from .shaper import TopicShaper


class TopicDataEntityHelper:
	"""
	use topic id as entity name
	"""

	def __init__(self, schema: TopicSchema):
		self.schema = schema
		self.shaper = self.create_entity_shaper(schema)
		self.entityHelper = EntityHelper(name=self.schema.get_topic().topicId, shaper=self.shaper)
		self.entityIdHelper = EntityIdHelper(
			name=self.schema.get_topic().topicId,
			shaper=self.shaper,
			idColumnName=TopicDataColumnNames.ID.value
		)

	def get_schema(self) -> TopicSchema:
		return self.schema

	def get_topic(self) -> Topic:
		return self.schema.get_topic()

	def get_entity_name(self) -> str:
		return self.get_entity_helper().name

	@abstractmethod
	def create_entity_shaper(self, schema: TopicSchema) -> TopicShaper:
		pass

	def get_entity_shaper(self) -> TopicShaper:
		return self.shaper

	def get_column_name(self, factor_name: str) -> Optional[str]:
		return self.shaper.get_mapper().get_column_name(factor_name)

	def get_factor_name(self, column_name: str) -> Optional[str]:
		return self.shaper.get_mapper().get_factor_name(column_name)

	def get_entity_helper(self) -> EntityHelper:
		return self.entityHelper

	def serialize_to_storage(self, data: Dict[str, Any]) -> Dict[str, Any]:
		return self.get_entity_helper().shaper.serialize(data)

	def deserialize_to_memory(self, data: Dict[str, Any]) -> Dict[str, Any]:
		return self.get_entity_helper().shaper.deserialize(data)

	def get_entity_id_helper(self) -> EntityIdHelper:
		return self.entityIdHelper

	def get_entity_finder(self, criteria: EntityCriteria, sort: Optional[EntitySort] = None) -> EntityFinder:
		entity_helper = self.get_entity_helper()
		return EntityFinder(
			name=entity_helper.name,
			shaper=entity_helper.shaper,
			criteria=criteria,
			sort=sort
		)

	def get_entity_pager(
			self, criteria: EntityCriteria, pageable: Pageable, sort: Optional[EntitySort] = None) -> EntityPager:
		entity_helper = self.get_entity_helper()
		return EntityPager(
			name=entity_helper.name,
			shaper=entity_helper.shaper,
			criteria=criteria,
			sort=sort,
			pageable=pageable
		)

	def get_distinct_values_finder(
			self,
			criteria: Optional[EntityCriteria], column_names: List[EntityColumnName],
			sort: Optional[EntitySort] = None,
			distinct_value_on_single_column: bool = False) -> EntityDistinctValuesFinder:
		entity_helper = self.get_entity_helper()
		return EntityDistinctValuesFinder(
			name=entity_helper.name,
			shaper=entity_helper.shaper,
			criteria=criteria,
			sort=sort,
			distinctColumnNames=column_names,
			distinctValueOnSingleColumn=distinct_value_on_single_column
		)

	def get_straight_values_finder(
			self,
			criteria: EntityCriteria, columns: List[EntityStraightColumn],
			sort: Optional[EntitySort] = None) -> EntityStraightValuesFinder:
		entity_helper = self.get_entity_helper()
		return EntityStraightValuesFinder(
			name=entity_helper.name,
			shaper=entity_helper.shaper,
			criteria=criteria,
			sort=sort,
			straightColumns=columns
		)

	def get_entity_updater(self, criteria: EntityCriteria, update: EntityUpdate) -> EntityUpdater:
		entity_helper = self.get_entity_helper()
		return EntityUpdater(
			name=entity_helper.name,
			shaper=entity_helper.shaper,
			criteria=criteria,
			update=update
		)

	def get_entity_deleter(self, criteria: EntityCriteria) -> EntityDeleter:
		entity_helper = self.get_entity_helper()
		return EntityDeleter(
			name=entity_helper.name,
			shaper=entity_helper.shaper,
			criteria=criteria
		)

	# noinspection PyMethodMayBeStatic
	def find_data_id(self, data: Dict[str, Any]) -> Tuple[bool, Optional[int]]:
		"""
		find data if from given data dictionary.
		"""
		id_ = data.get(TopicDataColumnNames.ID.value)
		return id_ is not None, id_

	def build_id_criteria(self, data: Dict[str, Any]) -> Optional[EntityCriteriaExpression]:
		"""
		return none when id not found
		"""
		has_id, id_ = self.find_data_id(data)
		if not has_id:
			return None
		else:
			return EntityCriteriaExpression(left=ColumnNameLiteral(columnName=TopicDataColumnNames.ID.value), right=id_)

	# noinspection PyMethodMayBeStatic
	def assign_id_column(self, data: Dict[str, Any], id_value: int) -> None:
		data[TopicDataColumnNames.ID.value] = id_value

	@abstractmethod
	def is_versioned(self) -> bool:
		pass

	@abstractmethod
	def find_version(self, data: Dict[str, Any]) -> int:
		pass

	@abstractmethod
	def build_version_criteria(self, data: Dict[str, Any]) -> Optional[EntityCriteriaExpression]:
		"""
		return none when topic is not versioned or version not found
		"""
		pass

	@abstractmethod
	def assign_version(self, data: Dict[str, Any], version: int) -> None:
		"""
		default ignore version assignment
		"""
		pass

	# noinspection PyMethodMayBeStatic
	def assign_tenant_id(self, data: Dict[str, Any], tenant_id: TenantId) -> None:
		data[TopicDataColumnNames.TENANT_ID.value] = tenant_id

	# noinspection PyMethodMayBeStatic
	def find_insert_time(self, data: Dict[str, Any]) -> Optional[datetime]:
		return data.get(TopicDataColumnNames.INSERT_TIME.value)

	# noinspection PyMethodMayBeStatic
	def assign_insert_time(self, data: Dict[str, Any], insert_time: datetime) -> None:
		data[TopicDataColumnNames.INSERT_TIME.value] = insert_time

	# noinspection PyMethodMayBeStatic
	def find_update_time(self, data: Dict[str, Any]) -> Optional[datetime]:
		return data.get(TopicDataColumnNames.UPDATE_TIME.value)

	# noinspection PyMethodMayBeStatic
	def assign_update_time(self, data: Dict[str, Any], update_time: datetime) -> None:
		data[TopicDataColumnNames.UPDATE_TIME.value] = update_time

	def assign_fix_columns_on_create(
			self, data: Dict[str, Any],
			snowflake_generator: SnowflakeGenerator, principal_service: PrincipalService,
			now: datetime
	) -> None:
		self.assign_id_column(data, snowflake_generator.next_id())
		self.assign_tenant_id(data, principal_service.get_tenant_id())
		self.assign_insert_time(data, now)
		self.assign_update_time(data, now)
		self.assign_version(data, 1)

	def assign_fix_columns_on_update(
			self, data: Dict[str, Any],
			principal_service: PrincipalService,
			now: datetime, version: int
	) -> None:
		self.assign_tenant_id(data, principal_service.get_tenant_id())
		self.assign_update_time(data, now)
		self.assign_version(data, version)



from abc import abstractmethod
from logging import getLogger
from typing import Any, Callable, Dict, Optional

from watchmen_data_kernel.common import DataKernelException
from watchmen_model.common import DataModel

logger = getLogger(__name__)


class ExternalWriterParams(DataModel):
	pat: Optional[str] = None
	url: Optional[str] = None
	eventCode: Optional[str] = None
	currentData: Optional[Dict[str, Any]] = None
	previousData: Optional[Dict[str, Any]] = None
	variables: Optional[Dict[str, Any]] = None


class ExternalWriter:
	def __init__(self, code: str):
		self.code = code

	def get_code(self) -> str:
		"""
		code should be same as configured
		"""
		return self.code

	@abstractmethod
	def run(self, params: ExternalWriterParams) -> bool:
		pass


BuildExternalWriter = Callable[[str], ExternalWriter]


class ExternalWriterRegistry:
	builders: Dict[str, BuildExternalWriter] = {}

	def register(self, code: str, build_writer: BuildExternalWriter) -> Optional[BuildExternalWriter]:
		original = self.builders.get(code)
		self.builders[code] = build_writer
		if original is not None:
			logger.warning(f'External writer[{code}] is replaced.')
		else:
			logger.info(f'External writer[{code}] is registered.')
		return original

	def is_registered(self, code: str) -> bool:
		return code in self.builders

	def ask_writer(self, code: str) -> BuildExternalWriter:
		create = self.builders.get(code)
		if create is None:
			raise DataKernelException(f'Creator not found for external writer[{code}].')
		return create


external_writer_registry = ExternalWriterRegistry()


def register_external_writer_creator(code: str, create_writer: BuildExternalWriter) -> None:
	"""
	register writer on startup
	"""
	external_writer_registry.register(code, create_writer)


def is_external_writer_creator_registered(code: str) -> bool:
	return external_writer_registry.is_registered(code)


def ask_external_writer_creator(code: str) -> BuildExternalWriter:
	return external_writer_registry.ask_writer(code)



from typing import Optional

from watchmen_auth import PrincipalService
from watchmen_data_kernel.common import ask_replace_topic_to_storage, ask_sync_topic_to_storage, ask_trino_enabled
from watchmen_model.admin import Topic, TopicKind
from watchmen_utilities import is_blank
from .storage_helper import ask_topic_storage


def beautify_name(topic: Topic) -> str:
	return '' if is_blank(topic.name) else topic.name.strip().lower()


def can_do_next(topic: Topic) -> bool:
	if is_blank(topic.dataSourceId):
		return False
	if beautify_name(topic) == '':
		return False
	return True


def create_topic_structure(topic: Topic, principal_service: PrincipalService) -> None:
	if not can_do_next(topic):
		return

	storage = ask_topic_storage(topic, principal_service)
	# create topic only when topic is not a synonym
	if topic.kind != TopicKind.SYNONYM:
		storage.create_topic_entity(topic)


def drop_topic_structure(topic: Topic, principal_service: PrincipalService) -> None:
	if not can_do_next(topic):
		return

	storage = ask_topic_storage(topic, principal_service)
	# drop topic only when topic is not a synonym
	if topic.kind != TopicKind.SYNONYM:
		storage.drop_topic_entity(topic)


def update_topic_structure(topic: Topic, original_topic: Topic, principal_service: PrincipalService) -> None:
	if not can_do_next(topic):
		return

	storage = ask_topic_storage(topic, principal_service)
	# update topic structure only when topic is not a synonym
	if topic.kind != TopicKind.SYNONYM:
		storage.update_topic_entity(topic, original_topic)


def sync_for_trino(topic: Topic, original_topic: Optional[Topic], principal_service: PrincipalService) -> None:
	if original_topic is not None and can_do_next(original_topic):
		storage = ask_topic_storage(original_topic, principal_service)
		storage.drop_topic_from_trino(original_topic)

	if can_do_next(topic):
		storage = ask_topic_storage(topic, principal_service)
		storage.append_topic_to_trino(topic)


def sync_topic_structure_storage(
		topic: Topic, original_topic: Optional[Topic], principal_service: PrincipalService) -> None:
	if ask_trino_enabled():
		sync_for_trino(topic, original_topic, principal_service)

	if not ask_sync_topic_to_storage():
		return

	if original_topic is None:
		create_topic_structure(topic, principal_service)
	elif ask_replace_topic_to_storage():
		drop_topic_structure(original_topic, principal_service)
		create_topic_structure(topic, principal_service)
	elif topic.dataSourceId != original_topic.dataSourceId:
		# not in same data source, leave original as is
		create_topic_structure(topic, principal_service)
	elif is_blank(original_topic.dataSourceId):
		# no data source declared in original, typically no storage entity existing
		# simply do create for new one
		create_topic_structure(topic, principal_service)
	elif beautify_name(topic) != beautify_name(original_topic):
		# name changed, leave original as is
		create_topic_structure(topic, principal_service)
	else:
		# with same name, same data source, update it
		update_topic_structure(topic, original_topic, principal_service)



from watchmen_data_kernel.cache import CacheService
from watchmen_data_kernel.storage import RawTopicDataEntityHelper, RegularTopicDataEntityHelper, TopicDataEntityHelper
from watchmen_data_kernel.topic_schema import TopicSchema
from watchmen_model.admin import is_raw_topic


def ask_topic_data_entity_helper(schema: TopicSchema) -> TopicDataEntityHelper:
	"""
	ask topic data entity helper, from cache first
	"""
	data_entity_helper = CacheService.topic().get_entity_helper(schema.get_topic().topicId)
	if data_entity_helper is None:
		if is_raw_topic(schema.get_topic()):
			data_entity_helper = RawTopicDataEntityHelper(schema)
		else:
			data_entity_helper = RegularTopicDataEntityHelper(schema)
		CacheService.topic().put_entity_helper(data_entity_helper)
	return data_entity_helper



from typing import Any, Dict, List, Optional

from watchmen_data_kernel.topic_schema import TopicSchema
from watchmen_model.admin import Factor, is_aggregation_topic
from watchmen_model.pipeline_kernel import TopicDataColumnNames
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityRow, EntityShaper
from watchmen_utilities import ArrayHelper
from .data_entity_helper import TopicDataEntityHelper
from .data_service import TopicDataService
from .factor_column_mapper import TopicFactorColumnMapper
from .shaper import TopicShaper


class RegularTopicFactorColumnMapper(TopicFactorColumnMapper):
	def get_factors(self, schema: TopicSchema) -> List[Factor]:
		return ArrayHelper(schema.get_topic().factors).to_list()


class RegularTopicShaper(TopicShaper):
	def create_factor_column_mapper(self, schema: TopicSchema) -> TopicFactorColumnMapper:
		return RegularTopicFactorColumnMapper(schema)

	def serialize(self, data: Dict[str, Any]) -> EntityRow:
		if self.is_synonym():
			row: EntityRow = {}
		else:
			row = self.serialize_fix_columns(data)
			if is_aggregation_topic(self.get_schema().get_topic()):
				row[TopicDataColumnNames.AGGREGATE_ASSIST.value] = data.get(TopicDataColumnNames.AGGREGATE_ASSIST.value)
				row[TopicDataColumnNames.VERSION.value] = data.get(TopicDataColumnNames.VERSION.value)
		ArrayHelper(self.get_mapper().get_factor_names()).each(lambda x: self.serialize_factor(data, x, row))
		return row

	def deserialize(self, row: EntityRow) -> Dict[str, Any]:
		if self.is_synonym():
			data: Dict[str, Any] = {}
		else:
			data = self.deserialize_fix_columns(row)
			if is_aggregation_topic(self.get_schema().get_topic()):
				data[TopicDataColumnNames.AGGREGATE_ASSIST.value] = row.get(TopicDataColumnNames.AGGREGATE_ASSIST.value)
				data[TopicDataColumnNames.VERSION.value] = row.get(TopicDataColumnNames.VERSION.value)
		ArrayHelper(self.get_mapper().get_column_names()).each(lambda x: self.deserialize_column(row, x, data))
		return data


class RegularTopicDataEntityHelper(TopicDataEntityHelper):
	def create_entity_shaper(self, schema: TopicSchema) -> EntityShaper:
		return RegularTopicShaper(schema)

	def is_versioned(self) -> bool:
		return is_aggregation_topic(self.get_schema().get_topic())

	def find_version(self, data: Dict[str, Any]) -> int:
		if is_aggregation_topic(self.get_schema().get_topic()):
			return data.get(TopicDataColumnNames.VERSION.value)
		else:
			return -1

	def build_version_criteria(self, data: Dict[str, Any]) -> Optional[EntityCriteriaExpression]:
		return EntityCriteriaExpression(
			left=ColumnNameLiteral(columnName=TopicDataColumnNames.VERSION.value), right=self.find_version(data))

	def assign_version(self, data: Dict[str, Any], version: int):
		if is_aggregation_topic(self.get_schema().get_topic()):
			data[TopicDataColumnNames.VERSION.value] = version


class RegularTopicDataService(TopicDataService):
	def try_to_wrap_to_topic_data(self, data: Dict[str, Any]) -> Dict[str, Any]:
		self.delete_reversed_columns(data)
		return data

	def try_to_unwrap_from_topic_data(self, topic_data: Dict[str, Any]) -> Dict[str, Any]:
		return topic_data



from typing import List, Optional

from watchmen_data_kernel.storage import TopicDataEntityHelper
from watchmen_data_kernel.topic_schema import TopicSchema
from watchmen_model.admin import Topic
from watchmen_model.common import TenantId, TopicId
from .cache_manager import get_topic_by_id_cache, get_topic_by_tenant_and_name_cache, \
	get_topic_entity_helper_by_id_cache, get_topic_schema_by_id_cache
from .internal_cache import InternalCache
from .pipeline_by_topic_cache import pipeline_by_topic_cache


class TopicCache:
	def __init__(self):
		self.byIdCache = InternalCache(cache=get_topic_by_id_cache)
		self.byTenantAndNameCache = InternalCache(cache=get_topic_by_tenant_and_name_cache)
		self.schemaByIdCache = InternalCache(cache=get_topic_schema_by_id_cache)
		self.entityHelperByIdCache = InternalCache(cache=get_topic_entity_helper_by_id_cache)

	# noinspection PyMethodMayBeStatic
	def to_tenant_and_name_key(self, name: str, tenant_id: TenantId) -> str:
		return f'{tenant_id}-{name}'

	def put(self, topic: Topic) -> Optional[Topic]:
		# topic is changed, remove from entity helper cache anyway
		self.entityHelperByIdCache.remove(topic.topicId)
		# refresh other caches
		existing_topic = self.byIdCache.put(topic.topicId, topic)
		self.byTenantAndNameCache.put(
			self.to_tenant_and_name_key(topic.name, topic.tenantId), topic)
		self.schemaByIdCache.put(topic.topicId, TopicSchema(topic))
		return existing_topic

	def put_entity_helper(self, entity_helper: TopicDataEntityHelper) -> Optional[TopicDataEntityHelper]:
		return self.entityHelperByIdCache.put(entity_helper.get_topic().topicId, entity_helper)

	def get(self, topic_id: TopicId) -> Optional[Topic]:
		return self.byIdCache.get(topic_id)

	def get_schema(self, topic_id: TopicId) -> Optional[TopicSchema]:
		return self.schemaByIdCache.get(topic_id)

	def get_entity_helper(self, topic_id: TopicId) -> Optional[TopicDataEntityHelper]:
		return self.entityHelperByIdCache.get(topic_id)

	def get_by_name(self, name: str, tenant_id: TenantId) -> Optional[Topic]:
		return self.byTenantAndNameCache.get(self.to_tenant_and_name_key(name, tenant_id))

	def remove(self, topic_id: TopicId) -> Optional[Topic]:
		existing: Optional[Topic] = self.byIdCache.remove(topic_id)
		if existing is not None:
			pipeline_by_topic_cache.remove(topic_id)
			self.byTenantAndNameCache.remove(self.to_tenant_and_name_key(existing.name, existing.tenantId))
		self.schemaByIdCache.remove(topic_id)
		self.entityHelperByIdCache.remove(topic_id)
		return existing

	def all(self) -> List[Topic]:
		return list(self.byIdCache.values())

	def clear(self) -> None:
		self.byIdCache.clear()
		self.byTenantAndNameCache.clear()
		self.schemaByIdCache.clear()
		self.entityHelperByIdCache.clear()
		pipeline_by_topic_cache.clear()


topic_cache = TopicCache()



from datetime import date
from typing import Any, List, Union

from watchmen_data_kernel.common import DataKernelException
from watchmen_model.admin import FactorEncryptMethod
from watchmen_utilities import is_date_plus_format
from .encryptor import Encryptor


class DateMasker(Encryptor):
	def __init__(self, month: bool, day: bool, formats: List[str]):
		self.maskMonth = month
		self.maskDay = day
		if month and day:
			self.method = FactorEncryptMethod.MASK_MONTH_DAY
		elif month:
			self.method = FactorEncryptMethod.MASK_MONTH
		elif day:
			self.method = FactorEncryptMethod.MASK_DAY
		else:
			raise DataKernelException(f'At least one of month or day should be masked, current is none.')
		self.formats = formats
		if formats is None or len(formats) == 0:
			raise DataKernelException(f'At least one date format should be provided, current is none.')

	def accept(self, method: Union[FactorEncryptMethod, str]) -> bool:
		return method == self.method

	def is_encrypted(self, value: Any) -> bool:
		"""
		always returns false
		"""
		return False

	def do_encrypt(self, value: Any) -> Any:
		if isinstance(value, date):
			if self.method == FactorEncryptMethod.MASK_MONTH_DAY:
				return value.replace(month=1, day=1)
			elif self.method == FactorEncryptMethod.MASK_MONTH:
				return value.replace(month=1)
			elif self.method == FactorEncryptMethod.MASK_DAY:
				return value.replace(day=1)
			else:
				raise DataKernelException(f'Encrypt method[{self.method}] is not supported.')

		if not isinstance(value, str):
			raise DataKernelException(f'Given value[{value}] is not a string, cannot be masked.')

		parsed, date_value, date_format = is_date_plus_format(value, self.formats)
		if not parsed:
			raise DataKernelException(
				f'Given value[{value}] cannot be parsed to date or datetime by formats[{self.formats}].')
		# call myself to encrypt date
		masked_date = self.do_encrypt(date_value)
		# format to string again, now only decimal included
		formatted_value = masked_date.strftime(date_format)
		# test len, return directly when length is same
		if len(formatted_value) == len(value):
			return formatted_value

		# gather decimal character from formatted, other character from original
		final_value = []
		formatted_value_index = 0
		for ch in value:
			if not ch.isdecimal():
				final_value.append(ch)
			else:
				final_value.append(formatted_value[formatted_value_index])
				formatted_value_index = formatted_value_index + 1
		return ''.join(final_value)

	def do_decrypt(self, value: str) -> str:
		# center mask cannot be decrypted
		return value

	def get_key_type(self) -> str:
		return self.method.value



from typing import List, Optional

from watchmen_model.common import PipelineId, TopicId
from watchmen_utilities import ArrayHelper
from .cache_manager import get_pipeline_by_topic_id_cache
from .internal_cache import InternalCache


class PipelineByTopicCache:
	def __init__(self):
		self.byTopicIdCache = InternalCache(cache=get_pipeline_by_topic_id_cache)

	def get(self, topic_id: TopicId) -> Optional[List[PipelineId]]:
		"""
		none means not initialized. empty list means no pipeline.
		"""
		return self.byTopicIdCache.get(topic_id)

	def declare_no_pipelines(self, topic_id: TopicId) -> None:
		self.byTopicIdCache.put(topic_id, [])

	def append_one(self, topic_id: TopicId, pipeline_id: PipelineId) -> None:
		"""
		append given pipeline id into by topic id cache
		"""
		pipeline_ids: Optional[List[PipelineId]] = self.byTopicIdCache.get(topic_id)
		if pipeline_ids is None:
			self.byTopicIdCache.put(topic_id, [pipeline_id])
		elif pipeline_id not in pipeline_ids:
			self.byTopicIdCache.put(topic_id, ArrayHelper(pipeline_ids).grab(pipeline_id).to_list())

	def remove_one(self, topic_id: TopicId, pipeline_id: PipelineId) -> None:
		"""
		remove given pipeline id from by topic id cache
		"""
		pipeline_ids: Optional[List[PipelineId]] = self.byTopicIdCache.get(topic_id)
		if pipeline_ids is not None:
			remains = ArrayHelper(pipeline_ids).filter(lambda x: x != pipeline_id).to_list()
			if len(remains) == 0:
				self.byTopicIdCache.remove(topic_id)
			else:
				self.byTopicIdCache.put(topic_id, remains)

	def remove(self, topic_id: TopicId) -> None:
		self.byTopicIdCache.remove(topic_id)

	def clear(self):
		self.byTopicIdCache.clear()


pipeline_by_topic_cache = PipelineByTopicCache()



from typing import Optional

from watchmen_auth import PrincipalService
from watchmen_data_kernel.cache import CacheService
from watchmen_meta.common import ask_meta_storage
from watchmen_meta.system import KeyStoreService as KeyStoreStorageService
from watchmen_model.common import TenantId
from watchmen_model.system import KeyStore


class KeyStoreService:
	def __init__(self, principal_service: PrincipalService):
		self.principalService = principal_service

	# noinspection PyMethodMayBeStatic
	def find_by_type(self, key_type: str, tenant_id: TenantId) -> Optional[KeyStore]:
		hit, key_store = CacheService.key_store().get(key_type, tenant_id)
		if hit:
			return key_store

		storage_service = KeyStoreStorageService(ask_meta_storage())
		storage_service.begin_transaction()
		try:
			key_store: KeyStore = storage_service.find_by_type(key_type, tenant_id)
			if key_store is None:
				CacheService.key_store().declare_not_existing(key_type, tenant_id)
				return None

			CacheService.key_store().put(key_store)
			return key_store
		finally:
			storage_service.close_transaction()



from __future__ import annotations

from datetime import date, time
from typing import Any, Dict, List, Optional, Union

from watchmen_data_kernel.common import ask_all_date_formats, ask_time_formats
from watchmen_data_kernel.common.settings import ask_abandon_date_time_on_parse_fail
from watchmen_model.admin import Factor, FactorType, Topic
from watchmen_utilities import ArrayHelper, is_blank, is_not_blank, try_to_date, try_to_time

"""
design of date/time/datetime translator. same with encryption, see that for more. 
"""


class DateOrTimeFactor:
	def __init__(self, factor: Factor):
		self.factor = factor
		self.factorName = '' if is_blank(factor.name) else factor.name.strip()
		self.names = self.factorName.split('.')

	def pop_first_name(self):
		self.names = self.names[1:]

	def translate(self, value: Any) -> Optional[Union[date, time, Any]]:
		factor_type = self.factor.type
		if factor_type == FactorType.FULL_DATETIME:
			parsed_value = try_to_date(value, ask_all_date_formats())
		elif factor_type == FactorType.DATETIME:
			parsed_value = try_to_date(value, ask_all_date_formats())
		elif factor_type == FactorType.DATE or factor_type == FactorType.DATE_OF_BIRTH:
			parsed_value = try_to_date(value, ask_all_date_formats())
		elif factor_type == FactorType.TIME:
			parsed_value = try_to_time(value, ask_time_formats())
		else:
			parsed_value = value
		if parsed_value is not None:
			return parsed_value
		elif ask_abandon_date_time_on_parse_fail():
			return None
		else:
			return value


class DateOrTimeFactorGroup:
	factors: Optional[List[DateOrTimeFactor]] = None
	groups: Optional[List[DateOrTimeFactorGroup]] = None

	def __init__(self, name: str, factors: List[DateOrTimeFactor]):
		self.name = name
		# in reality, zero or one factor.
		# if there is one, name is same as group's, and will not contain group anymore
		self.factors = ArrayHelper(factors).filter(lambda x: len(x.names) == 1).to_list()
		groups = ArrayHelper(factors).filter(lambda x: len(x.names) > 1) \
			.each(lambda x: x.pop_first_name()) \
			.group_by(lambda x: x.names[0])
		self.groups = ArrayHelper(list(groups.items())) \
			.map(lambda x: DateOrTimeFactorGroup(name=x[0], factors=x[1])).to_list()

	def translate(self, data: Dict[str, Any]) -> None:
		value = data.get(self.name)
		if value is None:
			return
		if isinstance(value, dict):
			ArrayHelper(self.groups).each(lambda x: x.translate(value))
		elif isinstance(value, list):
			def each(item):
				ArrayHelper(self.groups).each(lambda x: x.translate(item))

			ArrayHelper(value).each(lambda x: each(x))
		else:
			data[self.name] = self.factors[0].translate(value)


def is_date_or_time(factor: Factor) -> bool:
	factor_type = factor.type
	return factor_type in [
		FactorType.FULL_DATETIME, FactorType.DATETIME, FactorType.DATE, FactorType.TIME, FactorType.DATE_OF_BIRTH]


def parse_date_or_time_factors(topic: Topic) -> List[DateOrTimeFactorGroup]:
	groups = ArrayHelper(topic.factors) \
		.filter(is_date_or_time) \
		.map(lambda x: DateOrTimeFactor(x)) \
		.filter(lambda x: is_not_blank(x.factorName)) \
		.group_by(lambda x: x.names[0])

	return ArrayHelper(list(groups.items())) \
		.map(lambda x: DateOrTimeFactorGroup(name=x[0], factors=x[1])).to_list()
